#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains manager to handle tools
"""

from __future__ import print_function, division, absolute_import

import webbrowser

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc
from tpDcc.libs.python import color
from tpDcc.libs.qt.core import qtutils
from tpDcc.libs.qt.widgets import stack, buttons


class ToolsetWidget(stack.StackItem, object):

    ID = 'toolsetID'
    CONFIG = None
    EXTENSION = 'toolset'

    StartLargets = -1
    StartSmallest = 0

    displaySwitched = Signal()
    updatedPropertyRequested = Signal()
    savePropertyRequested = Signal()
    windowClosed = Signal()
    toolsetClosed = Signal()
    toolsetHidden = Signal()
    toolsetShown = Signal()
    toolsetDragged = Signal()
    toolsetDropped = Signal()
    toolsetDragCancelled = Signal()
    toolsetActivated = Signal()
    toolsetDeactivated = Signal()

    def __init__(self, tree_widget=None, icon_color=(255, 255, 255), widget_item=None, parent=None, *args, **kwargs):

        self._block_save = False
        self._show_warnings = True
        self._icon = self.CONFIG.get('icon') if self.CONFIG else None
        self._icon_color = icon_color
        self._tree_widget = tree_widget
        self._toolset_widget_item = widget_item
        self._stacked_widget = None
        self._display_mode_button = None
        self._help_button = None
        self._settings_button = None
        self._widgets = list()
        self._attacher = None
        title = self.CONFIG.data.get('label', '') if self.CONFIG else ''
        collapsable = kwargs.get('collapsable', True)
        show_item_icon = kwargs.get('show_item_icon', True)

        super(ToolsetWidget, self).__init__(
            title=title, collapsed=True, icon=self._icon, shift_arrows_enabled=False, title_editable=False,
            parent=parent or tree_widget, collapsable=collapsable, show_item_icon=show_item_icon,
            delete_button_enabled=True if tree_widget else False,
        )

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    @property
    def stacked_widget(self):
        return self._stacked_widget

    @property
    def attacher(self):
        return self._attacher

    # =================================================================================================================
    # TO OVERRIDE
    # =================================================================================================================

    def pre_content_setup(self):
        """
        Function that is called before toolset contents are created
        Override in specific toolset widgets
        """

        pass

    def contents(self):
        """
        Returns toolset widget contentS
        Override in specific toolset widgets
        :return:
        """

        pass

    def post_content_setup(self):
        """
        Function that is called after toolset contents are created
        Override in specific toolset widgets
        """

        pass

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def ui(self):
        super(ToolsetWidget, self).ui()

        self.setContentsMargins(0, 0, 0, 0)
        self._contents_layout.setContentsMargins(0, 0, 0, 0)
        self._contents_layout.setSpacing(0)

        self._stacked_widget = QStackedWidget(self._widget_hider)
        self._stacked_widget.setContentsMargins(0, 0, 0, 0)
        self._stacked_widget.setLineWidth(0)
        self._contents_layout.addWidget(self._stacked_widget)

        self.show_expand_indicator(False)
        self.set_title_text_mouse_transparent(True)

        self._display_mode_button = DisplayModeButton(color=self._icon_color, size=16, parent=self)
        self._display_mode_button.setFixedSize(QSize(22, 22))
        self._help_button = buttons.BaseMenuButton(parent=self)
        self._help_button.set_icon(tpDcc.ResourcesMgr().icon('help'))
        self._help_button.setFixedSize(QSize(22, 22))
        self._settings_button = buttons.BaseMenuButton(parent=self)
        self._settings_button.set_icon(tpDcc.ResourcesMgr().icon('settings'))
        self._settings_button.setFixedSize(QSize(22, 22))

        # TODO: Disable until we implement preferences system
        self._settings_button.setVisible(False)

        # We call if after setting all buttons
        self.set_icon_color(self._icon_color)

        self.visual_update(collapse=True)

        display_button_pos = 7
        self._title_frame.horizontal_layout.insertWidget(display_button_pos - 1, self._help_button)
        self._title_frame.horizontal_layout.insertWidget(display_button_pos - 2, self._settings_button)
        self._title_frame.horizontal_layout.insertWidget(display_button_pos, self._display_mode_button)
        self._title_frame.horizontal_layout.setSpacing(0)
        self._title_frame.horizontal_layout.setContentsMargins(0, 0, 0, 0)
        self._title_frame.delete_button.setIconSize(QSize(12, 12))
        self._title_frame.item_icon_button.setIconSize(QSize(20, 20))

        font = QFont()
        font.setBold(True)
        self.setFont(font)

    def setup_signals(self):
        super(ToolsetWidget, self).setup_signals()

        self._title_frame.mouseReleaseEvent = self._on_activate_event

        if self._toolset_widget_item:
            self._display_mode_button.clicked.connect(self._toolset_widget_item.set_current_index)
            self._display_mode_button.clicked.connect(lambda: self.displaySwitched.emit())
            self.displaySwitched.connect(lambda: self.updateRequested.emit())

        self._help_button.leftClicked.connect(self._on_open_help)
        self._settings_button.leftClicked.connect(self._on_show_preferences_dialog)

        self.deletePressed.connect(self._on_stop_callbacks)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_attacher(self, attacher):
        """
        Sets attacher of this toolset
        The attacher is the window this toolset belongs to. Can be a custom Qt window or Hub
        :param attacher:
        """

        self._attacher = attacher

    def get_icon(self):
        """
        Returns toolset icon
        :return: QIcon
        """

        if not self._title_frame:
            return tpDcc.ResourcesMgr().icon('tpdcc')

        return tpDcc.ResourcesMgr().icon(self._title_frame.item_icon or 'tpdcc')

    def widgets(self):
        """
        Returns all display widgets
        :return: list
        """

        return self._widgets

    def display_index(self):
        """
        Current index of the stacked widget
        :return: int
        """

        return self._stacked_widget.currentIndex()

    def current_widget(self):
        """
        Returns current active widget
        :return: QWidget
        """

        return self.widgets()[self.display_index()]

    def widget_at(self, index):
        """
        Returns stacked widget at given index
        :param index: int
        :return: QWidget
        """

        return self._stacked_widget.widget(index)

    def count(self):
        """
        Returns the total amount of widgets added to the stack
        :return: int
        """

        return self._stacked_widget.count()

    def add_stacked_widget(self, widget):
        """
        Adds a new widget to the stack
        :param widget: QWidget
        """

        if not widget:
            raise ValueError(
                'Toolset "{}" contents() must return a list of widgets! None found.'.format(
                    str(self.__class__.__name__)))

        self._widgets.append(widget)
        widget.setProperty('color', self._icon_color)
        self._stacked_widget.addWidget(widget)

    def visual_update(self, collapse=True):
        """
        Updates visual based on opened or closed state
        :param collapse: bool
        """

        if collapse:
            self.set_icon_color(color.desaturate(self._icon_color, 0.75), set_color=False)
            self.title_text_widget().setObjectName('disabled')
        else:
            self.set_icon_color(self._icon_color)
            self.title_text_widget().setObjectName('active')

        qtutils.update_widget_style(self.title_text_widget())

        self.setUpdatesEnabled(False)
        self.updatedPropertyRequested.emit()
        self.setUpdatesEnabled(True)

    def update_display_button(self):
        """
        Updates the display button based on number of widgets in the stack
        """

        self.set_displays(self.count())

    def set_displays(self, displays):
        """
        Sets displays
        :param displays: list
        """

        if displays in [ToolsetDisplays.Single, ToolsetDisplays.Double, ToolsetDisplays.Triple]:
            self._display_mode_button.set_displays(displays)
        else:
            tpDcc.logger.error('setDisplays() must be 2 or 3')

    def block_save(self, flag):
        """
        Sets whether data saving is enabled or not
        :param flag: bool
        """

        self._block_save = flag

    def set_active(self, active=True, emit=True):
        """
        Sets whether toolset widget is active or not
        :param active: bool
        :param emit: bool
        """

        if active:
            self.expand(emit)
        else:
            self.collapse(emit)
        self.visual_update(collapse=not active)

    def set_current_index(self, index):

        self.block_save(True)

        for i in range(self._stacked_widget.count()):
            w = self._stacked_widget.widget(i)
            w.setSizePolicy(w.sizePolicy().horizontalPolicy(), QSizePolicy.Ignored)

        self._stacked_widget.setCurrentIndex(index)

        widget = self._stacked_widget.widget(index)
        if widget:
            widget.setSizePolicy(widget.sizePolicy().horizontalPolicy(), QSizePolicy.Expanding)
        else:
            tpDcc.logger.warning('Widget not found!')

        self.block_save(False)

    def set_icon_color(self, color=(255, 255, 255), set_color=True):
        """
        Sets the icon color for all the icons
        :param color: tuple(int, int, int), color in RGB integer values in 0-255 range
        :param set_color: bool, saves current color overwriting cache variable
        """

        if set_color:
            self._icon_color = color
        darken = 0.8
        self.set_item_icon_color(color)
        self._display_mode_button.set_icon_color(color)
        self._help_button.set_icon_color((color[0] * darken, color[1] * darken, color[2] * darken))
        self._title_frame.delete_button.set_icon_color(color)

    def populate_widgets(self):

        property_widgets = self.property_widgets()

        # self.savePropertyRequested.connect(self.save_properties)
        # self.updatedPropertyRequested.connect(self.update_properties)

    # def linked_properties(self, widget):
    #     """
    #     Returns list of linkable properties from widgets children
    #     :param widget:  list
    #     :return:
    #     """
    #
    #     for attr in widget.__dict__:
    #         if type(getattr(widget, attr)) in
    #
    def auto_link_properties(self, widgets):
        """
        Add link properties if allowed
        :param widgets: list
        """

        if not self.CONFIG and self.CONFIG.get('auto_link_properties', True):
            return

        new_properties = list()
        names = list()
        # for link_property in self._linked

    def property_widgets(self, widget=None, current_widget=False):
        """
        Get properties widgets from the stack
        :param widget: QWidget
        :param current_widget: bool, Set True to only search the current active widget in the stack
        :return:
        """

        result = list()

        if current_widget:
            widget = widget or self._stacked_widget.current_widget()
        else:
            widget = widget or self._stacked_widget

        for child in qtutils.iterate_children(widget, skip='skipChildren'):
            if child.property('prop') is not None:
                result.append(child)

        return result

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _stop_selection_callback(self):
        print('Stop Selection Callbacks ...')

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_activate_event(self, event, emit=True):
        self._on_toggle_contents(emit=emit)
        event.ignore()

    def _on_open_help(self):
        """
        Internal callback function that is called when help button is clicked
        """

        url = self.CONFIG.get('help_url', '') if self.CONFIG else ''
        if not url:
            return
        webbrowser.open(url)

    def _on_stop_callbacks(self):
        self._stop_selection_callback()

    def _on_show_preferences_dialog(self):

        from tpDcc.libs.qt.widgets import lightbox
        from tpDcc.libs.qt.core import preferences
        self._lightbox = lightbox.Lightbox(self)
        self._lightbox.closed.connect(self._on_close_lightbox)
        self._preferences_window = preferences.PreferencesWidget(settings=self.preferences_settings())
        self._preferences_window.setFixedHeight(500)
        self._preferences_window.closed.connect(self._on_close_preferences_window)
        self._lightbox.set_widget(self._preferences_window)
        for pref_widget in self._preference_widgets_classes:
            pref_widget = pref_widget()
            self._preferences_window.add_category(pref_widget.CATEGORY, pref_widget)
        self._theme_widget = self._setup_theme_preferences()
        self._preferences_window.add_category(self._theme_widget.CATEGORY, self._theme_widget)
        self._lightbox.show()

    def _on_close_preferences_window(self, save_widget=False):
        self._on_close_lightbox(save_widget)
        self._lightbox.blockSignals(True)
        self._lightbox.close()
        self._lightbox.blockSignals(False)

    def _on_close_lightbox(self, save_widgets=False):
        if not save_widgets:
            self._settings_accepted(**self._theme_widget._dlg._form_widget.default_values())
        else:
            self._settings_accepted(**self._theme_widget._dlg._form_widget.values())


class ToolsetDisplays(object):
    """
    Display modes for the toolset widget items
    """

    Single = 1
    Double = 2
    Triple = 3


class DisplayModeButton(buttons.BaseMenuButton, object):

    FIRST_INDEX = 1
    LAST_INDEX = -1

    clicked = Signal(object)

    def __init__(self, parent=None, size=16, color=(255, 255, 255), initial_index=FIRST_INDEX):
        super(DisplayModeButton, self).__init__(parent=parent)

        menu_icon_double_names = ['menu_double_empty', 'menu_double_one', 'menu_double_full']
        menu_icon_triple_names = ['menu_triple_empty', 'menu_triple_one', 'menu_triple_two', 'menu_triple_full']
        self._menu_icon_double = [tpDcc.ResourcesMgr().icon(menu_icon) for menu_icon in menu_icon_double_names]
        self._menu_icon_triple = [tpDcc.ResourcesMgr().icon(menu_icon) for menu_icon in menu_icon_triple_names]

        self._current_icon = None
        self._icons = None
        self._displays = None
        self._initial_display = initial_index
        self._current_size = size
        self._icon_color = color
        self._highlight_offset = 40

        self.setIconSize(QSize(size, size))
        self.set_displays(ToolsetDisplays.Double)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def mouseReleaseEvent(self, event):
        """
        Overrides base QWidget mouseReleaseEvent
        :param event: QMouseEvent
        """

        new_icon, new_index = self.next_icon()
        self.set_icon(new_icon, size=self._current_size, colors=self._icon_color)
        self._current_icon = new_icon
        self.clicked.emit(new_index - 1)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_displays(self, displays=ToolsetDisplays.Triple):
        """
        Sets the number of displays
        :param displays:
        """

        self._displays = displays
        self.show()

        if displays == 3:
            self._icons = self._menu_icon_triple
        elif displays == 2:
            self._icons = self._menu_icon_double
        elif displays == 1:
            self.hide()
        else:
            tpDcc.logger.error('only 2 or 3 displays are available!')

        self._current_icon = self._icons[self._initial_display]
        self.set_icon_index(self._initial_display)

    def set_icon_index(self, index, size=None, color=None):
        """
        Sets icon by its display index
        :param index: int
        :param size:
        :param color:
        """

        if size is None:
            size = self._current_size
        if color is None:
            color = self._icon_color

        self.set_icon(self._icons[index], size=size, colors=color)

    def next_icon(self):
        """
        Returns next icon of the current one and its index
        :return: tuple(QIcon, int)
        """

        new_index = max((self._icons.index(self._current_icon) + 1) % len(self._icons), 1)
        new_icon = self._icons[new_index]

        return new_icon, new_index



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for page widgets
"""

from __future__ import print_function, division, absolute_import

import math
from functools import partial

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.qt.core import mixin, base, menu, theme
from tpDcc.libs.qt.widgets import label, buttons, combobox, spinbox


@mixin.theme_mixin
class Page(base.BaseWidget, mixin.FieldMixin):
    """
    Widget that allows to divide long list of items in several pages.
    Only one page will be loaded at a time
    """

    pageChanged = Signal(int, int)

    def __init__(self, parent=None):
        super(Page, self).__init__(parent=parent)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def get_main_layout(self):
        main_layout = QHBoxLayout()
        main_layout.setSpacing(2)
        main_layout.setContentsMargins(0, 0, 0, 0)

        return main_layout

    def ui(self):
        super(Page, self).ui()

        current_theme = self.theme()

        self._page_setting_menu = menu.BaseMenu(parent=self)

        self._display_label = label.BaseLabel()
        self._display_label.setAlignment(Qt.AlignCenter)
        self._change_page_size_btn = combobox.BaseComboBox().small()
        self._change_page_size_btn.setFixedWidth(110)
        self._change_page_size_btn.set_menu(self._page_setting_menu)
        self._change_page_size_btn.set_formatter(lambda x: '{} per page'.format(x))

        self._pre_btn = buttons.BaseToolButton().image('arrow_left').icon_only().small()
        self._next_btn = buttons.BaseToolButton().image('arrow_right').icon_only().small()
        self._current_page_spin_box = spinbox.BaseSpinBox()
        self._current_page_spin_box.setMinimum(1)
        self._current_page_spin_box.theme_size = current_theme.small if current_theme else theme.Theme.Sizes.SMALL
        self._total_page_lbl = label.BaseLabel()

        self.main_layout.addStretch()
        self.main_layout.addWidget(self._display_label)
        self.main_layout.addStretch()
        self.main_layout.addWidget(label.BaseLabel('|').secondary())
        self.main_layout.addWidget(self._change_page_size_btn)
        self.main_layout.addWidget(label.BaseLabel('|').secondary())
        self.main_layout.addWidget(self._pre_btn)
        self.main_layout.addWidget(label.BaseLabel('Page'))
        self.main_layout.addWidget(self._current_page_spin_box)
        self.main_layout.addWidget(label.BaseLabel('/'))
        self.main_layout.addWidget(self._total_page_lbl)
        self.main_layout.addWidget(self._next_btn)

    def setup_signals(self):

        self.register_field('page_size_selected', 25)
        self.register_field('page_size_list', [
            {'label': '25 - Fastest', 'value': 25},
            {'label': '50 - Fast', 'value': 50},
            {'label': '75 - Medium', 'value': 75},
            {'label': '100 - Slow', 'value': 100},
        ])
        self.register_field('total', 0)
        self.register_field('current_page', 0)
        self.register_field('total_page', lambda: self._get_total_pages(
            self.field('total'), self.field('page_size_selected')))
        self.register_field('total_page_text', lambda: str(self.field('total_page')))
        self.register_field('display_text', lambda: self._get_page_display_string(
            self.field('current_page'), self.field('page_size_selected'), self.field('total')))
        self.register_field('can_pre', lambda: self.field('current_page') > 1)
        self.register_field('can_next', lambda: self.field('current_page') < self.field('total_page'))

        self._change_page_size_btn.valueChanged.connect(self._on_page_changed)
        self._current_page_spin_box.valueChanged.connect(self._on_page_changed)
        self._pre_btn.clicked.connect(partial(self._on_change_current_page, -1))
        self._next_btn.clicked.connect(partial(self._on_change_current_page, 1))

        self.bind('page_size_list', self._page_setting_menu, 'data')
        self.bind('page_size_selected', self._page_setting_menu, 'value', signal='valueChanged')
        self.bind('page_size_selected', self._change_page_size_btn, 'value', signal='valueChanged')
        self.bind('current_page', self._current_page_spin_box, 'value', signal='valueChanged')
        self.bind('total_page', self._current_page_spin_box, 'maximum')
        self.bind('total_page_text', self._total_page_lbl, 'label_text')
        self.bind('display_text', self._display_label, 'label_text')
        self.bind('can_pre', self._pre_btn, 'enabled')
        self.bind('can_next', self._next_btn, 'enabled')

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_total(self, value):
        """
        Sets the total amount of pages
        :param value: int
        """

        self.set_field('total', value)
        self.set_field('current_page', 1)

    def set_page_config(self, data_list):
        """
        Sets page component per page settings
        :param data_list: list(dict)
        """

        self.set_field('page_size_list', [
            {'label': str(data), 'value': data} if isinstance(data, int) else data for data in data_list])

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _get_total_pages(self, total, count_per_page):
        """
        Internal function that returns the total amount of pages available
        :param total: int
        :param count_per_page: int
        :return: int
        """

        return int(math.ceil(1.0 * total / count_per_page))

    def _get_page_display_string(self, current, count_per_page, total):
        """
        Returns the page string in proper format (x - x of xx)
        :param current: int
        :param count_per_page: int
        :param total: int
        :return: str
        """

        return '{} - {} of {}'.format(
            (current - 1) * count_per_page + 1 if current else 0, min(total, current * count_per_page), total)

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_change_current_page(self, offset):
        """
        Internal callback function that is called when a page is changed
        :param offset: int
        """

        self.set_field('current_page', self.field('current_page') + offset)
        self._on_page_changed()

    def _on_page_changed(self):
        """
        Internal callback function that is called when a page is changed. Emits pageChanged signal.
        """

        self.pageChanged.emit(self.field('page_size_selected'), self.field('current_page'))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains functions related with Qt dock behaviour
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs.qt.widgets import drop

USE_ANIMATIONS = True


# ===================================================================================================================

def new_dock_splitter(orientation=Qt.Horizontal, parent=None):
    s = QSplitter(orientation, parent)
    s.setProperty('dock-splitter', True)
    s.setChildrenCollapsible(False)
    s.setOpaqueResize(False)

    return s


def splitter_contains_section_widget(splitter):
    for i in range(splitter.count()):
        w = splitter.widget(i)
        if splitter and splitter_contains_section_widget(splitter):
            return True
        else:
            if issubclass(w, DockSectionWidget):
                return True

    return False


def delete_empty_splitter(container):
    do_again = False
    while do_again:
        do_again = False
        splitters = container.findChildren(QSplitter)
        for i in range(len(splitters)):
            sp = splitters.at(i)
            if not sp.property('dock-splitter').toBool():
                continue
            if sp.countT() > 0 and splitter_contains_section_widget(sp):
                continue

            del splitters[i]
            do_again = True
            break


def find_parent_container_widget(widget):
    next_widget = widget
    while next_widget:
        if isinstance(next_widget, DockContainer):
            break
        next_widget = next_widget.parentWidget()

    return next_widget


def find_parent_section_widget(widget):
    next_widget = widget
    while next_widget:
        if isinstance(next_widget, DockSectionWidget):
            break
        next_widget = next_widget.parentWidget()

    return next_widget


def find_parent_splitter(widget):
    next_widget = widget
    while next_widget:
        if isinstance(next_widget, QSplitter):
            break
        next_widget = next_widget.parentWidget()

    return next_widget


def find_inmediate_splitter(widget):
    sp = None
    layout = widget.layout()
    if not layout or layout.count() <= 0:
        return sp

    for i in range(layout.count()):
        layout_item = layout.itemAt(0)
        if not layout_item.widget():
            continue
        if isinstance(layout_item.widget(), QSplitter):
            sp = layout_item.widget()
            break

    return sp


# ===================================================================================================================

class DockInternalContentData(object):
    content = None
    title_widget = None
    content_widget = None


class DockHiddenSectionItem(object):
    def __init__(self):
        self.preferred_section_id = -1
        self.preferred_section_index = -1
        self.data = DockInternalContentData()


class DockFlags(object):
    NoFlag = 0
    Closeable = 1
    AllFlags = Closeable


class DockSectionContent(object):

    next_uid = 0

    def __init__(self):
        super(DockSectionContent, self).__init__()

        self._uid = self.get_next_uid()
        self._unique_name = ''
        self._title = ''
        self._flags = [DockFlags.AllFlags]
        self._container_widget = None
        self._content_widget = None
        self._title_widget = None

        tp.logger.debug('Creating DockSectionContent {}'.format(self._uid))

    def get_next_uid(self):
        next_id = DockSectionContent.next_uid
        DockSectionContent.next_uid += 1
        return next_id

    def uid(self):
        return self._uid

    def unique_name(self):
        return self._unique_name

    def container_widget(self):
        return self._container_widget

    def content_widget(self):
        return self._content_widget

    def title_widget(self):
        return self._title_widget

    def flags(self):
        return self._flags

    def title(self):
        return self._title

    def set_title(self, title):
        self._title = title

    def set_flags(self, flags):
        self._flags = flags

    def clean(self):
        if self._container_widget:
            del self._container_widget._section_container_lookup_map_by_id[self._uid]
            del self._container_widget._section_container_lookup_map_by_name[self._unique_name]
            self._title_widget.deleteLater()
            del self._title_widget
            self._content_widget.deleteLater()
            del self._content_widget

    @staticmethod
    def new_section_content(unique_name, container, title, content):
        """
        Creates a new DockSectionWidget an register it into its DockContainer
        :param unique_name: str, unique name for the DockSectionWidget
        :param container: DockContainer, container where DockSectionWidget is stored
        :param title: str, title of the DockSectionWidget
        :param content: QWidget, widget that DockSectionWidget should store and manage
        :return: QDockSectionWidget
        """

        if unique_name == '':
            tp.logger.error('Cannot create DockSectionContent with empty unique_name')
            return None
        elif DockContainer.check_section_container_lookup_map_by_name(container, unique_name):
            tp.logger.error('Cannot create DockSectionContent with already used unique_name')
            return None
        elif not container or not title or not content:
            tp.logger.error('Cannot create DockSectionContent with None values')
            return None

        section_content = DockSectionContent()
        section_content._unique_name = unique_name
        section_content._container_widget = container
        section_content._title_widget = QLabel(title)
        section_content._content_widget = content
        container.register_section_content(section_content)

        return section_content

    def visible_title(self):
        if self._title == '':
            return self._unique_name

        return self._title


class DockSectionContentWidget(QFrame, object):
    def __init__(self, section_content, parent=None):
        super(DockSectionContentWidget, self).__init__(parent=parent)

        self._content = section_content

        self.main_layout = QBoxLayout(QBoxLayout.TopToBottom)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.main_layout.addWidget(self._content.content_widget())
        self.setLayout(self.main_layout)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def clean(self):
        self.main_layout.removeWidget(self._content.content_widget())


class DockSectionTitleWidget(QFrame, object):

    clicked = Signal()
    activeTabChanged = Signal(bool)

    def __init__(self, section_content, parent=None):
        super(DockSectionTitleWidget, self).__init__(parent=parent)

        self._content = section_content
        self._tab_moving = False
        self._active_tab = False
        self._drag_start_pos = QPoint()
        self._floating_widget = None

        self.main_layout = QBoxLayout(QBoxLayout.LeftToRight)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.main_layout.addWidget(section_content.title_widget())
        self.setLayout(self.main_layout)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            event.accept()
            self._drag_start_pos = event.pos()
            tp.logger.debug('Starting to drag Dock from pos: {}'.format(self._drag_start_pos))
            return

        super(DockSectionTitleWidget, self).mousePressEvent(event)

    def _on_anim_move(self, data, container_widget, section_widget=None, loc=None):
        if section_widget:
            move_anim = QPropertyAnimation(self._floating_widget, 'pos', self)
            move_anim.setStartValue(self._floating_widget.pos())
            move_anim.setEndValue(section_widget.mapToGlobal(section_widget.rect().topLeft()))
            move_anim.setDuration(150)
            resize_anim = QPropertyAnimation(self._floating_widget, 'size', self)
            resize_anim.setStartValue(self._floating_widget.size())
            resize_anim.setEndValue(section_widget.size())
            resize_anim.setDuration(150)
            anim_grp = QParallelAnimationGroup(self)
            anim_grp.finished.connect(lambda: self._on_move_widget(data, container_widget, section_widget, loc))
            anim_grp.addAnimation(move_anim)
            anim_grp.addAnimation(resize_anim)
            anim_grp.start(QAbstractAnimation.DeleteWhenStopped)
        else:
            self._on_move_widget(data, container_widget, section_widget, loc)

    def _on_move_widget(self, data, container_widget, section_widget=None, loc=None):
        data = self._floating_widget.take_content(data)
        self._floating_widget.deleteLater()
        self._floating_widget.clean()
        self._floating_widget = None
        if loc and section_widget:
            container_widget.drop_content(data, section_widget, loc, True)

    def mouseReleaseEvent(self, event):
        container_widget = find_parent_container_widget(widget=self)

        # If we are dragging a widget
        if self._floating_widget:

            # We detect if we are over a DockContentWidget
            section_widget = container_widget.section_at(container_widget.mapFromGlobal(event.globalPos()))
            if section_widget:
                container_widget.drop_overlay().set_allowed_areas(drop.DropArea.AllAreas)
                loc = container_widget.drop_overlay().show_drop_overlay(section_widget)

                # If the drop is in an invalid area we put by default the container in the center
                if loc == drop.DropArea.InvalidDropArea:
                    loc = drop.DropArea.CenterDropArea

                if loc != drop.DropArea.InvalidDropArea:
                    data = DockInternalContentData()
                    if USE_ANIMATIONS:
                        self._on_anim_move(data, container_widget, section_widget, loc)
                    else:
                        self._on_move_widget(data, container_widget, section_widget, loc)
            else:
                drop_area = drop.DropArea.InvalidDropArea
                if container_widget.outer_top_drop_rect().contains(container_widget.mapFromGlobal(event.globalPos())):
                    drop_area = drop.DropArea.TopDropArea
                if container_widget.outer_right_drop_rect().contains(container_widget.mapFromGlobal(event.globalPos())):
                    drop_area = drop.DropArea.RightDropArea
                if container_widget.outer_bottom_drop_rect().contains(
                        container_widget.mapFromGlobal(event.globalPos())):
                    drop_area = drop.DropArea.BottomDropArea
                if container_widget.outer_left_drop_rect().contains(container_widget.mapFromGlobal(event.globalPos())):
                    drop_area = drop.DropArea.LeftDropArea

                # If the drop is in an invalid area we put by default the container in the center
                if drop_area == drop.DropArea.InvalidDropArea:
                    drop_area = drop.DropArea.CenterDropArea

                if drop_area != drop.DropArea.InvalidDropArea:
                    data = DockInternalContentData()
                    if USE_ANIMATIONS:
                        self._on_anim_move(data, container_widget)
                    else:
                        self._on_move_widget(data, container_widget)

                    container_widget.drop_content(data, None, drop_area, True)
        elif self._tab_moving and find_parent_section_widget(self) is not None:
            section = find_parent_section_widget(widget=self)
            pos = event.globalPos()
            pos = section.mapFromGlobal(pos)
            from_index = section.index_of_content(self._content)
            to_index = section.index_of_content_by_title_pos(pos, self)
            section.move_content(from_index, to_index)

        if not self._drag_start_pos.isNull():
            self.clicked.emit()

        self._drag_start_pos = QPoint()
        self._tab_moving = False
        container_widget.drop_overlay().hide_drop_overlay()
        super(DockSectionTitleWidget, self).mouseReleaseEvent(event)

    def mouseMoveEvent(self, event):

        # Get the DockContainer where this title is stored
        container_widget = find_parent_container_widget(widget=self)
        if container_widget is None:
            tp.logger.warning('Title is not stored in DockContainer! This should never happen!')
            return

        # Get the DockSectionWidget where this title is stored
        section_widget = find_parent_section_widget(widget=self)

        if self._floating_widget and (event.buttons() & Qt.LeftButton):
            event.accept()
            move_to_pos = event.globalPos() - (self._drag_start_pos + QPoint(7, 7))
            self._floating_widget.move(move_to_pos)

            # Mouse is over a section widget
            section = container_widget.section_at(container_widget.mapFromGlobal(QCursor.pos()))
            if section:
                container_widget.drop_overlay().set_allowed_areas(drop.DropArea.AllAreas)
                container_widget.drop_overlay().show_drop_overlay(section)

            # Mouse is at the edge of the Container Widget
            elif container_widget.outer_top_drop_rect().contains(container_widget.mapFromGlobal(QCursor.pos())):
                container_widget.drop_overlay().set_allowed_areas([drop.DropArea.TopDropArea])
                container_widget.drop_overlay().show_drop_overlay(
                    container_widget, container_widget.outer_top_drop_rect())
            elif container_widget.outer_right_drop_rect().contains(container_widget.mapFromGlobal(QCursor.pos())):
                container_widget.drop_overlay().set_allowed_areas([drop.DropArea.RightDropArea])
                container_widget.drop_overlay().show_drop_overlay(
                    container_widget, container_widget.outer_right_drop_rect())
            elif container_widget.outer_bottom_drop_rect().contains(container_widget.mapFromGlobal(QCursor.pos())):
                container_widget.drop_overlay().set_allowed_areas([drop.DropArea.BottomDropArea])
                container_widget.drop_overlay().show_drop_overlay(
                    container_widget, container_widget.outer_bottom_drop_rect())
            elif container_widget.outer_left_drop_rect().contains(container_widget.mapFromGlobal(QCursor.pos())):
                container_widget.drop_overlay().set_allowed_areas([drop.DropArea.LeftDropArea])
                container_widget.drop_overlay().show_drop_overlay(
                    container_widget, container_widget.outer_left_drop_rect())
            else:
                container_widget.drop_overlay().hide_drop_overlay()
            return
        elif not self._floating_widget and not self._drag_start_pos.isNull() and (
                event.buttons() & Qt.LeftButton) and section_widget is not None and \
                not section_widget.title_area_geometry().contains(section_widget.mapFromGlobal(event.globalPos())):
            event.accept()

            # We get the new dock info for the current section
            # DockSectionTitleWidget and DockSectionContentWidgets are reparented to the DockContainer and are ready
            # to be moved to a new DockSectionWidget
            tp.logger.debug('Creating new floating dock data ...')
            data = section_widget.take_content(self._content.uid())
            if data is None:
                tp.logger.error(
                    'This should not happen! {} - {}'.format(self._content.uid(), self._content.unique_name()))
                return

            # Create floating widget and add it to the list of Container floatters
            tp.logger.debug('Creating Floating Widget ...')
            self._floating_widget = DockFloatingWidget(container=container_widget, section_content=data.content,
                                                       section_title_widget=data.title_widget,
                                                       content_widget=data.content_widget, parent=container_widget)
            self._floating_widget.resize(section_widget.size())
            container_widget.floatings().append(self._floating_widget)

            move_to_pos = event.globalPos() - (self._drag_start_pos + QPoint(7, 7))
            self._floating_widget.move(move_to_pos)
            self._floating_widget.show()

            # Delete old section, if it is empty now
            if section_widget.is_empty():
                section_widget.clean()
                section_widget.deleteLater()
                del section_widget

            delete_empty_splitter(container=container_widget)
            return
        elif self._tab_moving and find_parent_section_widget(self):
            event.accept()
            move_to_pos = self.mapToParent(event.pos()) - self._drag_start_pos
            move_to_pos.setY(0)
            self.move(move_to_pos)
            return
        elif not self._drag_start_pos.isNull() and (
                event.buttons() & Qt.LeftButton) and (event.pos() - self._drag_start_pos).manhattanLength() >= \
                QApplication.startDragDistance() and find_parent_section_widget(self) is not None \
                and find_parent_section_widget(self).title_area_geometry().contains(
                find_parent_section_widget(self).mapFromGlobal(event.globalPos())):
            event.accept()
            self._tab_moving = True
            self.raise_()
            return

        super(DockSectionTitleWidget, self).mouseMoveEvent(event)

    def clean(self):
        self.main_layout.removeWidget(self._content.title_widget())

    def content(self):
        return self._content

    def is_active_tab(self):
        return self._active_tab

    def set_active_tab(self, active):
        if active != self._active_tab:
            self._active_tab = active
            self.style().unpolish(self)
            self.style().polish(self)
            self.update()
            self.activeTabChanged.emit(active)

    def _on_anim_finished(self, content_widget, data, section_widget, loc):
        data = self._floating_widget.take_content(data)
        self._floating_widget.clean()
        self._floating_widget.deleteLater()
        self._floating_widget = None
        content_widget.drop_content(data, section_widget, loc)


class DockSectionWidgetTabsScrollArea(QScrollArea, object):
    """
    Custom scrollable implementation for tabs
    """

    def __init__(self, section_widget, parent=None):
        super(DockSectionWidgetTabsScrollArea, self).__init__(parent=parent)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setFrameStyle(QFrame.NoFrame)
        self.setWidgetResizable(True)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.setMaximumHeight(15)

    def wheelEvent(self, event):
        event.accept()
        try:
            # For Qt >= 5.0.0
            direction = event.angleDelta().y()
        except Exception:
            direction = event.delta()

        if direction < 0:
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() + 20)
        else:
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() - 20)


class DockSectionWidget(QFrame, object):

    next_uid = 0

    def __init__(self, container_widget=None):
        super(DockSectionWidget, self).__init__(parent=container_widget)

        self._uid = self.get_next_uid()
        self._container = container_widget
        self._tabs_layout_init_count = 0
        self._mouse_press_title_widget = None

        self._contents = list()
        self._section_titles = list()
        self._section_contents = list()

        self.main_layout = QBoxLayout(QBoxLayout.TopToBottom)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.setLayout(self.main_layout)

        self._top_layout = QBoxLayout(QBoxLayout.LeftToRight)
        self._top_layout.setContentsMargins(0, 0, 0, 0)
        self._top_layout.setSpacing(0)
        self.main_layout.addLayout(self._top_layout)

        self._tabs_scroll_area = DockSectionWidgetTabsScrollArea(section_widget=self)
        self._top_layout.addWidget(self._tabs_scroll_area, 1)

        self._tabs_container_widget = QWidget()
        self._tabs_container_widget.setObjectName('tabsContainerWidget')
        self._tabs_scroll_area.setWidget(self._tabs_container_widget)

        self._tabs_layout = QBoxLayout(QBoxLayout.LeftToRight)
        self._tabs_layout.setContentsMargins(0, 0, 0, 0)
        self._tabs_layout.setSpacing(0)
        self._tabs_layout.addStretch(1)
        self._tabs_container_widget.setLayout(self._tabs_layout)

        self._tabs_menu_button = QPushButton()
        self._tabs_menu_button.setObjectName('tabsMenuButton')
        self._tabs_menu_button.setFlat(True)
        self._tabs_menu_button.setIcon(self.style().standardIcon(QStyle.SP_TitleBarUnshadeButton))
        self._tabs_menu_button.setMaximumWidth(self._tabs_menu_button.iconSize().width())
        self._top_layout.addWidget(self._tabs_menu_button, 0)

        self._close_btn = QPushButton()
        self._close_btn.setObjectName('closeButton')
        self._close_btn.setFlat(True)
        self._close_btn.setIcon(self.style().standardIcon(QStyle.SP_TitleBarCloseButton))
        self._close_btn.setMaximumWidth(self._close_btn.iconSize().width())
        self._top_layout.addWidget(self._close_btn, 0)
        self._close_btn.clicked.connect(self._on_close_button_clicked)

        self._tabs_layout_init_count = self._tabs_layout.count()

        # Central area with contents
        self._contents_layout = QStackedLayout()
        self._contents_layout.setContentsMargins(0, 0, 0, 0)
        self._contents_layout.setSpacing(0)
        self.main_layout.addLayout(self._contents_layout, 1)

        shadow_effect = QGraphicsDropShadowEffect(self)
        shadow_effect.setOffset(0, 0)
        shadow_effect.setBlurRadius(8)
        self.setGraphicsEffect(shadow_effect)

    @staticmethod
    def get_next_uid():
        next_id = DockSectionWidget.next_uid
        DockSectionWidget.next_uid += 1
        return next_id

    def uid(self):
        return self._uid

    def container_widget(self):
        return self._container

    def contents(self):
        return self._contents

    def is_empty(self):
        return len(self._contents) <= 0

    def showEvent(self, event):
        """
        When the DockSectionWidget is showed, we make need to make sure that the widget is visible
        :param event: QShowEvent
        """

        try:
            self._tabs_scroll_area.ensureWidgetVisible(self._section_titles[self.current_index()])
        except Exception:
            pass

    def title_area_geometry(self):
        return self._top_layout.geometry()

    def content_area_geometry(self):
        return self._contents_layout.geometry()

    def clean(self):
        if self._container:
            del self._container._section_widget_lookup_map_by_id[self._uid]
            self._container._sections.remove(self)

        splitter = find_parent_splitter(self)
        if splitter and splitter.count() == 0:
            splitter.deleteLater()
            del splitter

    def current_index(self):
        """
        Returns the widget index that QStackedWidget is showing right now
        :return: int
        """

        return self._contents_layout.currentIndex()

    def set_current_index(self, index):
        if index < 0 or index > len(self._contents) - 1:
            tp.logger.warning('Invalid index: {}'.format(index))
            return

        for i in range(self._tabs_layout.count()):
            item = self._tabs_layout.itemAt(i)
            if item.widget():
                section_title_widet = item.widget()
                if isinstance(section_title_widet, DockSectionTitleWidget):
                    if i == index:
                        section_title_widet.set_active_tab(True)
                        self._tabs_scroll_area.ensureWidgetVisible(section_title_widet)
                        if DockFlags.Closeable in section_title_widet._content.flags():
                            self._close_btn.setEnabled(True)
                        else:
                            self._close_btn.setEnabled(False)
                    else:
                        section_title_widet.set_active_tab(False)

        # Set active content
        self._contents_layout.setCurrentIndex(index)

    def update_tabs_menu(self):
        m = QMenu()
        for i in range(len(self._contents)):
            section_content = self._contents[i]
            a = m.addAction(QIcon(), section_content.visible_title())
            a.setData(section_content.uid())
            a.triggered.connect(self._on_tabs_menu_action_triggered)

        old = self._tabs_menu_button.menu()
        self._tabs_menu_button.setMenu(m)

        if old:
            old.deleteLater()
            del old

    def add_content(self, section_content):
        """
        Creates DockSectionTitleWidget and DockSectionContentWidget from scratch and store on them
        the given DockSectionContent
        :param section_content: DockSectionContent
        """

        # Add the given DockSectionContent to the list of contents
        self._contents.append(section_content)

        # Create new DockSectionTitle and add it to the list of title widgets and to the layout
        title = DockSectionTitleWidget(section_content=section_content)
        self._section_titles.append(title)
        self._tabs_layout.insertWidget(self._tabs_layout.count() - self._tabs_layout_init_count, title)
        title.clicked.connect(self._on_section_title_clicked)

        # Create DockSectionContentWidget that will store the given DockSectionContent
        content = DockSectionContentWidget(section_content=section_content)
        self._section_contents.append(content)
        self._contents_layout.addWidget(content)

        # If we only have stored one DockSectionContent, we set the current index to the first one
        if len(self._contents) == 1:
            self.set_current_index(0)

        # Update the tab menu depending on the DockSectionContents number
        self.update_tabs_menu()

    def add_data_content(self, internal_data, auto_activate):
        """
        Adds DockSectionTitleWidget and DockSectionContentWidget from the given data. Also the DockSectionContent is
        retrieved from the data
        :param internal_data: DockInternalContentData
        :param auto_activate: bool
        """

        # Add the DockSectionContent stored in the data into the list of contents
        self._contents.append(internal_data.content)

        # Add the DockSectionTitle stored in the data to the list of title widgets and to the layout
        self._section_titles.append(internal_data.title_widget)
        self._tabs_layout.insertWidget(
            self._tabs_layout.count() - self._tabs_layout_init_count, internal_data.title_widget)
        internal_data.title_widget.clicked.connect(self._on_section_title_clicked)

        # Add DockSectionContentWidget stored in the data and associates the DockSectionContent stored also in the data
        self._section_contents.append(internal_data.content_widget)
        self._contents_layout.addWidget(internal_data.content_widget)

        if len(self._contents) == 1:
            self.set_current_index(0)
        elif auto_activate:
            self.set_current_index(len(self._contents) - 1)
        else:
            internal_data.title_widget.set_active_tab(False)

        # Update the tab menu depending on the DockSectionContents number
        self.update_tabs_menu()

    def take_content(self, uid):
        """

        :param uid: uid of the DockSectionContent we want to extract from this section
        :return: bool, whether the new data is valid or not
        """

        section_content = None
        index = -1

        # Check if the given UID is valid and is already stored in this DockSectionWidget
        # Once, is found, we store its index and we extract it from the list of DockSectionContents
        # If no valid DockSectionContent is found, we abort the process
        for i in range(len(self._contents)):
            if self._contents[i].uid() != uid:
                continue
            index = i
            section_content = self._contents.pop(i)
            break
        if section_content is None:
            return None

        # We get the DockSectionTitleWidget associated with the DockSectionContent and:
        # 1. Remove it from the DockSectionWidget layout
        # 2. Disconnect all its signals
        # 3. Parent it to the DockContainer
        title = self._section_titles.pop(index)
        if title:
            self._tabs_layout.removeWidget(title)
            # title.setAttribute(Qt.WA_WState_Created, False)
            title.disconnect(self)
            title.setParent(self._container)
            # title.setAttribute(Qt.WA_WState_Created, True)

        # We get the DockSectionContentWidget associated with the DockSectionContent and:
        # 1. Remove it from the DockSectionWidget
        # 2. Disconnect all its signals
        # 3. Parent it to the DockContainer
        content = self._section_contents.pop(index)
        if content:
            self._contents_layout.removeWidget(content)
            content.disconnect(self)
            content.setParent(self._container)

        # We update the active tab
        if len(self._contents) > 0 and title.is_active_tab():
            if index > 0:
                self.set_current_index(index - 1)
            else:
                self.set_current_index(0)

        # We update the tabs menu
        self.update_tabs_menu()

        if section_content is None or title is None or content is None:
            return None

        internal_data = DockInternalContentData()
        internal_data.content = section_content
        internal_data.title_widget = title
        internal_data.content_widget = content

        # try:
        #     return not internal_data.content.isNull()
        # except Exception:
        #     return internal_data.content is not None
        return internal_data

    def index_of_content(self, section_content):
        if section_content in self._contents:
            return self._contents.index(section_content)

    def index_of_content_by_uid(self, uid):
        for i in range(len(self._contents)):
            if self._contents[i].uid() == uid:
                return i

        return -1

    def index_of_content_by_title_pos(self, p, exclude):
        index = -1
        for i in range(len(self._section_titles)):
            if self._section_titles[i].geometry().contains(p) \
                    and (exclude is None or self._section_titles[i] != exclude):
                index = i
                break

        return index

    def move_content(self, from_index, to_index):
        if from_index >= len(self._contents) or from_index < 0 or \
                to_index >= len(self._contents) or to_index < 0 or from_index == to_index:
            tp.logger.warning('Invalid for tab movement - From: {} | To: {}'.format(from_index, to_index))
            self._tabs_layout.update()
            return

        self._contents.move(from_index, to_index)
        self._section_titles.move(from_index, to_index)
        self._section_contents.move(from_index, to_index)

        layout_from = self._tabs_layout.takeAt(from_index)
        self._tabs_layout.insertWidget(to_index, layout_from.widget())
        layout_from.deleteLater()
        del layout_from
        layout_from = None

        layout_from = self._contents_layout.takeAt(from_index)
        self._contents_layout.insertWidget(to_index, layout_from.widget())
        layout_from.deleteLater()
        del layout_from

        self.update_tabs_menu()

    def _on_close_button_clicked(self):
        index = self.current_index()
        if index < 0 or index > len(self._contents) - 1:
            return
        section_content = self._contents[index]
        if section_content is None:
            return
        self._container.remove_section_content(section_content)

    def _on_section_title_clicked(self):
        section_title_widget = self.sender()
        if not section_title_widget or not isinstance(section_title_widget, DockSectionTitleWidget):
            return

        index = self._tabs_layout.indexOf(section_title_widget)
        self.set_current_index(index)

    def _on_tabs_menu_action_triggered(self, flag):
        a = self.sender()
        if not a or not isinstance(a, QAction):
            return

        uid = a.data().toInt()
        index = self.index_of_content_by_uid(uid=uid)
        if index >= 0:
            self.set_current_index(index)


class DockFloatingWidget(QWidget, object):
    """
    Holds and displays DockSectionContent as a floating window
    It can be resized, moved and dropped back into a DockSectionWidget
    """

    def __init__(self, container, section_content, section_title_widget, content_widget, parent=None):
        super(DockFloatingWidget, self).__init__(parent=parent)

        self._container = container
        self._content = section_content
        self._title_widget = section_title_widget
        self._content_widget = content_widget

        self.main_layout = QBoxLayout(QBoxLayout.TopToBottom)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.setLayout(self.main_layout)

        self._title_layout = QBoxLayout(QBoxLayout.LeftToRight)
        self._title_layout.addWidget(section_title_widget, 1)
        self.main_layout.addLayout(self._title_layout, 0)
        section_title_widget.set_active_tab(False)

        if DockFlags.Closeable in section_content.flags():
            close_btn = QPushButton()
            close_btn.setObjectName('CloseButton')
            close_btn.setFlat(True)
            close_btn.setIcon(self.style().standardIcon(QStyle.SP_TitleBarCloseButton))
            close_btn.setToolTip('Close')
            close_btn.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
            self._title_layout.addWidget(close_btn)
            close_btn.clicked.connect(self._on_close_button_clicked)

        self.main_layout.addWidget(content_widget, 1)
        content_widget.show()

    def clean(self):
        del self._container._floatings[:]         # Python 2

    def take_content(self, data):
        data.content = self._content
        data.title_widget = self._title_widget
        data.content_widget = self._content_widget
        self._title_layout.removeWidget(self._title_widget)
        self._title_widget.setParent(self._container)
        self._title_widget = None
        self.main_layout.removeWidget(self._content_widget)
        self._content_widget.setParent(self._container)
        self._content_widget = None

        return data

    def _on_close_button_clicked(self):
        self._container.hide_section_content(self._content)


class DockContainer(QFrame, object):

    orientationChanged = Signal()
    activeTabChanged = Signal(object, bool)
    sectionContentVisibilityChanged = Signal(object, bool)

    def __init__(self, parent=None):
        super(DockContainer, self).__init__(parent=parent)

        self._orientation = Qt.Horizontal
        self._drop_overlay = drop.DropOverlay(parent=self)
        self._splitter = None
        self._sections = list()
        self._floatings = list()
        self._hidden_section_contents = dict()

        self._section_container_lookup_map_by_id = dict()
        self._section_container_lookup_map_by_name = dict()
        self._section_widget_lookup_map_by_id = dict()

        self.main_layout = QGridLayout()
        self.main_layout.setContentsMargins(9, 9, 9, 9)
        self.setLayout(self.main_layout)

    @staticmethod
    def check_section_container_lookup_map_by_id(dock_container, container_id):
        return container_id in dock_container._section_container_lookup_map_by_id

    @staticmethod
    def check_section_container_lookup_map_by_name(dock_container, name):
        return name in dock_container._section_container_lookup_map_by_name

    @staticmethod
    def check_section_widget_lookup_map_by_id(dock_container, widget_id):
        return widget_id in dock_container._section_widget_lookup_map_by_id

    def floatings(self):
        return self._floatings

    def is_empty(self):
        """
        Returns whether the container has sections stored in it or not
        :return: bool
        """

        return len(self._sections) <= 0

    def register_section_content(self, section_content):
        """
        Registers the given section content
        :param section_content: DockSectionContent
        """

        # TODO: We should check if the register is correct (no duplication unique names or ids)
        # TODO: And abort the process is something is wrong?

        self._section_container_lookup_map_by_id[section_content.uid()] = section_content
        self._section_container_lookup_map_by_name[section_content.unique_name()] = section_content

    def register_section_widget(self, section_widget):
        """
        Registers the given section widget
        :param section_widget: DockSectionWidget
        """

        # TODO: We should check if the register is correct (no duplication unique names or ids)
        # TODO: And abort the process is something is wrong?

        self._section_widget_lookup_map_by_id[section_widget.uid()] = section_widget
        self._sections.append(section_widget)

    def clean(self):
        self._sections.clear()
        del self._sections[:]

        self._floatings.clear()
        del self._floatings[:]

        self._sc_lookup_map_by_id.clear()
        self._sc_lookup_map_by_name.clear()
        self._sc_lookup_map_by_id.clear()

    def new_section_widget(self):
        """
        Creates a new DockSectionContentWidget from scratch
        :return: DockSectionContentWidget
        """

        section_widget = DockSectionWidget(container_widget=self)
        self.register_section_widget(section_widget)

        return section_widget

    def section_at(self, pos):
        """
        Returns the section, if exists, that is located on top of the cursor
        :param pos: QPoint
        :return: DockSectionWidget
        """

        g_pos = self.mapToGlobal(pos)
        for i in range(len(self._sections)):
            section_widget = self._sections[i]
            if section_widget.rect().contains(section_widget.mapFromGlobal(g_pos)):
                return section_widget

        return None

    def drop_content_outer_helper(self, parent_layout, internal_data, orientation, append):
        if parent_layout is None:
            return

        section_widget = self.new_section_widget()
        section_widget.add_data_content(internal_data=internal_data, auto_activate=True)

        old_splitter = find_inmediate_splitter(widget=self)
        if not old_splitter:
            sp = new_dock_splitter(orientation=orientation)
            if self.main_layout.count() > 0:
                tp.logger.warning('Still items in layout. This should never happen!')
                layout_item = self.main_layout.takeAt(0)
                layout_item.deleteLater()
                del layout_item

            self.main_layout.addWidget(sp)
            sp.addWidget(section_widget)
        elif old_splitter.orientation() == orientation or old_splitter.count() == 1:
            old_splitter.setOrientation(orientation)
            if append:
                old_splitter.addWidget(section_widget)
            else:
                old_splitter.insertWidget(0, section_widget)
        else:
            sp = new_dock_splitter(orientation=orientation)
            if append:
                try:
                    layout_item = self.main_layout.replaceWidget(old_splitter, sp)
                    sp.addWidget(old_splitter)
                    sp.addWidget(section_widget)
                    layout_item.deleteLater()
                    del layout_item
                except Exception:
                    index = self.main_layout.indexOf(old_splitter)
                    layout_item = self.main_layout.takeAt(index)
                    self.main_layout.addWidget(old_splitter)
                    self.main_layout.addWidget(section_widget)
                    layout_item.deleteLater()
                    del layout_item
            else:
                try:
                    sp.addWidget(section_widget)
                    item_layout = self.main_layout.replaceWidget(old_splitter, sp)
                    sp.addWidget(old_splitter)
                    item_layout.deleteLater()
                    del item_layout
                except Exception:
                    sp.addWidget(section_widget)
                    index = self.main_layout.indexOf(old_splitter)
                    item_layout = self.main_layout.takeAt(index)
                    self.main_layout.addWidget(sp)
                    sp.addWidget(old_splitter)
                    item_layout.deleteLater()
                    del item_layout

        return section_widget

    def drop_content(self, internal_data, target_section_widget, drop_area, auto_active):
        """
        Adds and fills DockSectionContent based on the given data and positioned in the given drop.DropArea
        :param internal_data: DockInternalContentData, stores all dock information to add to the DockContainer
        :param target_section_widget: DockSectionWidget
        :param drop_area: drop.DropArea, area where we want to drop the new DockSectionContent
        :param auto_active: bool, whether if the new DockSectionContent should be visible or not by default
        :return: DropSectionContent, new DropSectionContent added
        """

        ret = None

        # If the DockContainer has no DockSectionContents added to it, we create a new one from scratch and is loaded
        # automatically in the center of the DockContainer
        if self.is_empty():
            tp.logger.debug('DockContainer is empty! Creating DockSectionWidget from scratch ...')
            target_section_widget = self.new_section_widget()
            self.add_section(target_section_widget)
            drop_area = drop.DropArea.CenterDropArea

        # If no the container is not empty and we do not specify a DockSectionWidget, we create a new one
        # from scratch dropped in the given drop area
        if not target_section_widget:
            if drop_area == drop.DropArea.TopDropArea:
                ret = self.drop_content_outer_helper(
                    parent_layout=self.main_layout,
                    internal_data=internal_data, orientation=Qt.Vertical, append=False)
            elif drop_area == drop.DropArea.RightDropArea:
                ret = self.drop_content_outer_helper(
                    parent_layout=self.main_layout,
                    internal_data=internal_data, orientation=Qt.Horizontal, append=True)
            elif drop_area == drop.DropArea.CenterDropArea or drop_area == drop.DropArea.BottomDropArea:
                ret = self.drop_content_outer_helper(
                    parent_layout=self.main_layout,
                    internal_data=internal_data, orientation=Qt.Vertical, append=True)
            elif drop_area == drop.DropArea.LeftDropArea:
                self.drop_content_outer_helper(
                    parent_layout=self.main_layout,
                    internal_data=internal_data, orientation=Qt.Horizontal, append=False)
            else:
                return None
            return ret

        # We loop through all the widgets of the DockSectionWidget and we get its parent (splitter)
        target_section_splitter = find_parent_splitter(target_section_widget)
        if target_section_splitter is None:
            tp.logger.warning('DockSectionWidget is not parented to any widget. This cannot happen!')
            return

        # Add the DockSectionWidget into its corresponding area
        if drop_area == drop.DropArea.TopDropArea:
            section_widget = self.new_section_widget()
            section_widget.add_data_content(internal_data, True)
            if target_section_splitter.orientation() == Qt.Vertical:
                index = target_section_splitter.indexOf(target_section_widget)
                target_section_splitter.insertWidget(index, section_widget)
            else:
                index = target_section_splitter.indexOf(target_section_widget)
                s = new_dock_splitter(Qt.Vertical)
                s.addWidget(section_widget)
                s.addWidget(target_section_widget)
                target_section_splitter.insertWidget(index, s)
            ret = section_widget
        elif drop_area == drop.DropArea.RightDropArea:
            section_widget = self.new_section_widget()
            section_widget.add_data_content(internal_data, True)
            if target_section_splitter.orientation() == Qt.Horizontal:
                index = target_section_splitter.indexOf(target_section_widget)
                target_section_splitter.insertWidget(index + 1, section_widget)
            else:
                index = target_section_splitter.indexOf(target_section_widget)
                s = new_dock_splitter(Qt.Horizontal)
                s.addWidget(target_section_widget)
                s.addWidget(section_widget)
                target_section_splitter.insertWidget(index, s)
            ret = section_widget
        elif drop_area == drop.DropArea.BottomDropArea:
            section_widget = self.new_section_widget()
            section_widget.add_data_content(internal_data, True)
            if target_section_splitter.orientation() == Qt.Vertical:
                index = target_section_splitter.indexOf(target_section_widget)
                target_section_splitter.insertWidget(index + 1, section_widget)
            else:
                index = target_section_splitter.indexOf(target_section_widget)
                s = new_dock_splitter(Qt.Vertical)
                s.addWidget(target_section_widget)
                s.addWidget(section_widget)
                target_section_splitter.insertWidget(index, s)
            ret = section_widget
        elif drop_area == drop.DropArea.LeftDropArea:
            section_widget = self.new_section_widget()
            section_widget.add_data_content(internal_data, True)
            if target_section_splitter.orientation() == Qt.Horizontal:
                index = target_section_splitter.indexOf(target_section_widget)
                target_section_splitter.insertWidget(index, section_widget)
            else:
                s = new_dock_splitter(Qt.Horizontal)
                s.addWidget(section_widget)
                index = target_section_splitter.indexOf(target_section_widget)
                target_section_splitter.insertWidget(index, s)
                s.addWidget(target_section_widget)
            ret = section_widget
        elif drop_area == drop.DropArea.CenterDropArea:
            target_section_widget.add_data_content(internal_data, auto_activate=auto_active)
            ret = target_section_widget

        return ret

    def add_section(self, section_widget):
        """
        Adds the given DockSectionContentWidget to the Main Splitter DockContainer
        If the splitter does not exists, it's created automatically
        :param section_widget: DockSectionContentWidget
        """

        if not section_widget:
            tp.logger.warning('Impossible to add no valid DockSectionContentWidget!')
            return

        # If DockContainer's Main Splitter does not exists, we create it ...
        if not self._splitter:
            self._splitter = new_dock_splitter(orientation=self._orientation)
            self.main_layout.addWidget(self._splitter, 0, 0)

        # Check if the given DockSectionContentWidget has been already to the splitter.
        # If that's the case we do not add it
        if self._splitter.indexOf(section_widget) != -1:
            tp.logger.warning('Section has already been added!')
            return

        # If everything is ok, we add the DockSectionContentWidget to the splitter
        self._splitter.addWidget(section_widget)

    def take_content(self, section_content, internal_data):
        if section_content is None:
            return

        found = False

        # Search in sections
        for i in range(len(self._sections)):
            found = self._sections[i].take_content(section_content.uid(), internal_data)

        # Search in floating widgets
        if not found:
            for i in range(len(self._floatings)):
                found = self._floatings[i].content().uid() == section_content.uid()
                if found:
                    self._floatings[i].take_content(internal_data)

        # Search in hidden items
        if not found:
            if section_content.uid() in self._hidden_section_contents:
                hidden_section_item = self._hidden_section_contents[section_content.uid()]
                internal_data = hidden_section_item.data
                found = True

        return found

    def outer_top_drop_rect(self):
        r = self.rect()
        h = r.height() / 100 * 5
        return QRect(r.left(), r.top(), r.width(), h)

    def outer_right_drop_rect(self):
        r = self.rect()
        w = r.width() / 100 * 5
        return QRect(r.right() - w, r.top(), w, r.height())

    def outer_bottom_drop_rect(self):
        r = self.rect()
        w = r.width() / 100 * 5
        return QRect(r.right() - w, r.top(), w, r.height())

    def outer_left_drop_rect(self):
        r = self.rect()
        w = r.width() / 100 * 5
        return QRect(r.left(), r.top(), w, r.height())

    def contents(self):
        section_containers = self._section_container_lookup_map_by_id.values()
        contents = list()
        for i in range(len(section_containers)):
            section_content = section_containers[i]
            if section_content:
                contents.append(section_content)

        return contents

    def add_section_content(self, section_content, section_widget, drop_area):
        """
        Adds the given section content to this DockContainer
        If DockSectionWidget is not specified, a new one is created for the given DockSectionContent
        :param section_content: DockSectionContent we want to add to this DockContainer
        :param section_widget: DockSectionWidget that will be managed by the DockSectionContent
        :param drop_area: drop.DropArea, area where we want to add the new DockSectionContent into
        :return:
        """

        # If no DockSectionWidget is given, we abort the process
        if section_content is None:
            tp.logger.warning('Cannot add new DockSectionWidget if no one is specified')
            return

        # Create auxiliary data class to store dock data information
        # 1. DockSectionContent : stored from the given one
        # 2. Title Widget : created from scratch and parented to the DockSectionContent
        # 3. DockSectionContentWidget : create from scratch and parent to the DockSectionContent
        data = DockInternalContentData()
        data.content = section_content
        data.title_widget = DockSectionTitleWidget(section_content)
        data.content_widget = DockSectionContentWidget(section_content)

        # When the DOckSectionTitLeWidget is pressed  we notify that the tab have changed
        data.title_widget.activeTabChanged.connect(self._on_active_tab_changed)

        return self.drop_content(
            internal_data=data, target_section_widget=section_widget, drop_area=drop_area, auto_active=False)

    def show_section_content(self, section_content):
        if section_content is None:
            return

        # Search Section Content in floatings
        for i in range(len(self._floatings)):
            floating_widget = self._floatings[i]
            found = floating_widget.content().uid() == section_content.uid()
            if not found:
                continue
            floating_widget.setVisible(True)
            floating_widget._title_widget.setVisible(True)
            floating_widget._content_widget.setVisible(True)
            self.sectionContentVisibilityChanged.emit(section_content, True)
            return True

        # Search Section Content in hidden sections
        if section_content.uid() in self._hidden_section_contents:
            hidden_section_item = self._hidden_section_contents.pop(section_content.uid())
            hidden_section_item.data.title_widget.setVisible(True)
            hidden_section_item.data.contentWidget.setVisible(True)
            section_widget = None
            if hidden_section_item.preferred_section_id > 0:
                sw = self._section_container_lookup_map_by_id[hidden_section_item.preferred_section_id]
                if sw is not None:
                    sw.add_data_content(hidden_section_item.data, True)
                    self.sectionContentVisibilityChanged.emit(section_content, True)
                    return True
            elif len(self._sections) and self._sections[0] is not None:
                section_widget.add_data_content(hidden_section_item.data, True)
                self.sectionContentVisibilityChanged.emit(section_content, True)
                return True
            else:
                sw = self.new_section_widget()
                self.add_section(section_widget=sw)
                sw.add_data_content(hidden_section_item.data, True)
                self.sectionContentVisibilityChanged.emit(section_content, True)
                return True

        # Already visible?
        tp.logger.warning('Unable to show SectionContent, do not know where (already visible)?')
        return False

    def hide_section_content(self, section_content):
        """
        Hides the section content
        :param section_content: DockSectionContent you want to hide
        :return:
        """

        if section_content is None:
            return

        # Search Section Content in floatings
        for i in range(len(self._floatings)):
            found = self._floatings[i].content().uid() == section_content.uid()
            if not found:
                continue
            self._floatings[i].setVisible(False)
            self.sectionContentVisibilityChanged.emit(section_content, False)
            return True

        # Search Section Content in sections
        for i in range(len(self._sections)):
            section_widget = self._sections[i]
            found = section_widget.index_of_content(section_content) >= 0
            if not found:
                continue

            hidden_selection_item = DockHiddenSectionItem()
            hidden_selection_item.preferred_section_id = section_widget.uid()
            hidden_selection_item.preferred_section_index = section_widget.index_of_content(section_content)
            hidden_selection_item.data = section_widget.take_content(section_content.uid())
            if not hidden_selection_item.data:
                return False

            hidden_selection_item.data.title_widget.setVisible(False)
            hidden_selection_item.data.content_widget.setVisible(False)
            self._hidden_section_contents[section_content.uid()] = hidden_selection_item

            if section_widget.is_empty():
                section_widget.clean()
                section_widget.deleteLater()
                del section_widget
                delete_empty_splitter(self)

            self.sectionContentVisibilityChanged.emit(section_content, False)
            return True

        # Search Section Content in hidden elements
        if section_content.uid() in self._hidden_section_contents:
            return True

        tp.logger.error('Unable to hide SectionContent, do not know this one!')
        return False

    def raise_section_content(self, section_content):
        if section_content is None:
            return

        # Search Section Content in sections
        for i in range(len(self._sections)):
            sw = self._sections[i]
            index = sw.index_of_content(section_content)
            if index < 0:
                continue
            sw.set_current_index(index)
            return True

        # Search Section Content in floatins
        for i in range(len(self._floatings)):
            fw = self._floatings[i]
            if fw.content().uid() != section_content.uid():
                continue
            fw.setVisible(True)
            fw.raise_()
            return True

        # Search Section Content in hidden
        if section_content.uid() in self._hidden_section_contents:
            return self.show_section_content(section_content)

        tp.logger.error('Unable hide SectionContent, do not know this one!')
        return False

    def remove_section_content(self, section_content):
        if section_content is None:
            return

        if not self.hide_section_content(section_content):
            return False

        found = False
        for i in range(len(self._floatings)):
            fw = self._floatings[i]
            data = DockInternalContentData()
            found = fw.take_content(data)
            if not found:
                continue

            self._floatings.remove(fw)
            fw.clean()
            fw.deleteLater()
            del fw
            data.title_widget.clean()
            data.title_widget.deleteLater()
            del data.title_widget
            data.content_widget.clean()
            data.content_widget.deleteLater()
            del data.content_widget

        if section_content.uid() not in self._hidden_section_contents:
            tp.logger.error('Something went wrong ... The content should have been there!')
            return False

        hidden_selection_item = self._hidden_section_contents.pop(section_content.uid())
        hidden_selection_item.data.title_widget.deleteLater()
        del hidden_selection_item.data.title_widget
        hidden_selection_item.data.content_widget.deleteLater()
        del hidden_selection_item.data.content_widget

        try:
            section_content.clean()
            del section_content
        except Exception:
            tp.logger.warning('Maybe something went wrong when deleting section ...')

        if self.is_empty():
            self._splitter.deleteLater()
            del self._splitter
            self._splitter = None

        return True

    def is_section_content_visible(self, section_content):
        if section_content is None:
            return

        # Search Section Content in floatings
        for i in range(len(self._floatings)):
            found = self._floatings[i].content().uid() == section_content.uid()
            if not found:
                continue
            return self._floatings[i].isVisible()

        # Search Section Content in sections
        for i in range(len(self._sections)):
            sw = self._sections[i]
            index = sw.index_of_content(section_content)
            if index < 0:
                continue
            return True

        # Search Section Content in hidden
        if section_content.uid() in self._hidden_section_contents:
            return False

        tp.logger.warning(
            'SectionContent is not a part of this ContainerWidget: {}'.format(section_content.unique_name()))
        return False

    def drop_overlay(self):
        return self._drop_overlay

    def _on_active_tab_changed(self):
        section_title_widget = self.sender()
        if section_title_widget:
            self.activeTabChanged.emit(section_title_widget.content(), section_title_widget.is_active_tab())

    def _on_action_toggle_section_content_visibility(self, visible):
        a = self.sender()
        if not a:
            return
        uid = a.property('uid').toInt()
        section_content = self._section_container_lookup_map_by_id[uid]
        if not section_content:
            tp.logger.error('Cannot bind content by ID'.format(uid))
            return
        if visible:
            self.show_section_content(section_content)
        else:
            self.hide_section_content(section_content)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to create different kind of line edits
"""

from __future__ import print_function, division, absolute_import

from functools import partial

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.qt.core import mixin, theme
from tpDcc.libs.qt.widgets import layouts, buttons, browser


@mixin.theme_mixin
class BaseLineEdit(QLineEdit, object):
    """
     Basic line edit
     """

    delayTextChanged = Signal(str)

    def __init__(self, text='', parent=None):
        super(BaseLineEdit, self).__init__(text, parent)

        self._prefix_widget = None
        self._suffix_widget = None
        self._size = self.theme_default_size()

        self._main_layout = layouts.HorizontalLayout()
        self._main_layout.setContentsMargins(0, 0, 0, 0)
        self._main_layout.addStretch()
        self.setLayout(self._main_layout)

        self.setProperty('history', self.property('text'))
        self.setTextMargins(2, 0, 2, 0)

        self._delay_timer = QTimer()
        self._delay_timer.setInterval(500)
        self._delay_timer.setSingleShot(True)
        self._delay_timer.timeout.connect(self._on_delay_text_changed)

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    def _get_size(self):
        """
        Returns the spin box height size
        :return: float
        """

        return self._size

    def _set_size(self, value):
        """
        Sets spin box height size
        :param value: float
        """

        self._size = value
        if hasattr(self._prefix_widget, 'theme_size'):
            self._prefix_widget.theme_size = self._size
        if hasattr(self._suffix_widget, 'theme_size'):
            self._suffix_widget.theme_size = self._size
        self.style().polish(self)

    theme_size = Property(int, _get_size, _set_size)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def setText(self, text):
        """
        Overrides base QLineEdit setText base function.
        Save history
        :param text: str
        """

        self.setProperty('history', '{}\n{}'.format(self.property('history'), text))
        return super(BaseLineEdit, self).setText(text)

    def clear(self):
        """
        Overrides base QLineEdit clear function
        :return:
        """

        self.setProperty('history', '')
        return super(BaseLineEdit, self).clear()

    def keyPressEvent(self, event):
        """
        Overrides base QLineEdit keyPressEvent function
        :param event: QKeyEvent
        """

        if event.key() not in [Qt.Key_Enter, Qt.Key_Tab]:
            if self._delay_timer.isActive():
                self._delay_timer.stop()
            self._delay_timer.start()
        super(BaseLineEdit, self).keyPressEvent(event)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_delay_duration(self, ms):
        """
        Sets the delay timer duration
        :param ms: float
        """

        self._delay_timer.setInterval(ms)

    def get_prefix_widget(self):
        """
        Returns prefix widget for user to edit
        :return: QWidget
        """

        return self._prefix_widget

    def set_prefix_widget(self, widget):
        """
        Sets the edit line left start widget
        :param widget: QWidget
        :return: QWidget
        """

        if self._prefix_widget:
            index = self._main_layout.indexOf(self._prefix_widget)
            self._main_layout.takeAt(index)
            self._prefix_widget.setVisible(False)
            self._prefix_widget.deleteLater()

        widget.setProperty('combine', 'horizontal')
        widget.setProperty('position', 'left')
        if hasattr(widget, 'theme_size'):
            widget.them_size = self.theme_size

        margin = self.textMargins()
        margin.setLeft(margin.left() + widget.width())
        self.setTextMargins(margin)

        self._main_layout.insertWidget(0, widget)
        self._prefix_widget = widget

        return widget

    def get_suffix_widget(self):
        """
        Returns suffix widget for user to edit
        :return: QWidget
        """

        return self._suffix_widget

    def set_suffix_widget(self, widget):
        """
        Sets the edit line right start widget
        :param widget: QWidget
        :return: QWidget
        """

        if self._suffix_widget:
            index = self._main_layout.indexOf(self._suffix_widget)
            self._main_layout.takeAt(index)
            self._suffix_widget.setVisible(False)
            self._suffix_widget.deleteLater()

        widget.setProperty('combine', 'horizontal')
        widget.setProperty('position', 'right')
        if hasattr(widget, 'theme_size'):
            widget.them_size = self.theme_size

        margin = self.textMargins()
        margin.setRight(margin.right() + widget.width())
        self.setTextMargins(margin)

        self._main_layout.addWidget(widget)
        self._prefix_widget = widget

        return widget

    def search(self):
        """
        Adds a search icon button for line edit
        :return: self
        """

        prefix_btn = buttons.BaseToolButton().image('search').icon_only()
        suffix_btn = buttons.BaseToolButton().image('close').icon_only()
        suffix_btn.clicked.connect(self.clear)
        self.set_prefix_widget(prefix_btn)
        self.set_suffix_widget(suffix_btn)
        self.setPlaceholderText('Enter keyword to search ...')

        return self

    def search_engine(self, text='Search'):
        """
        Adds a search push button to line edit
        :param text: str
        :return: self
        """

        _suffix_btn = buttons.BaseButton(text).primary()
        _suffix_btn.clicked.connect(self.returnPressed)
        _suffix_btn.setFixedWidth(100)
        self.set_suffix_widget(_suffix_btn)
        self.setPlaceholderText('Enter keyword to search ...')

        return self

    def file(self, filters=None):
        """
        Adds a ClickBrowserFileToolButton to line edit
        :param filters:
        :return: self
        """

        _suffix_btn = browser.ClickBrowserFileToolButton()
        _suffix_btn.fileChanged.connect(self.setText)
        _suffix_btn.filters = filters
        self.textChanged.connect(_suffix_btn.set_path)
        self.set_suffix_widget(_suffix_btn)
        self.setPlaceholderText('Click button to browse files')

        return self

    def save_file(self, filters=None):
        """
        Adds a ClickSaveFileToolButton to line edit
        :param filters:
        :return: self
        """

        _suffix_button = browser.ClickSaveFileToolButton()
        _suffix_button.fileChanged.connect(self.setText)
        _suffix_button.filters = filters or list()
        self.textChanged.connect(_suffix_button.set_path)
        self.set_suffix_widget(_suffix_button)
        self.setPlaceholderText('Click button to set save file')

        return self

    def folder(self):
        """
        Adds a ClickBrowserFileToolButton to line edit
        :return: self
        """

        _suffix_btn = browser.ClickBrowserFolderToolButton()
        _suffix_btn.folderChanged.connect(self.setText)
        self.textChanged.connect(_suffix_btn.set_path)
        self.set_suffix_widget(_suffix_btn)
        self.setPlaceholderText('Click button to browse folder')

        return self

    def error(self):
        """
        Shows error in line edit with red style
        :return: self
        """

        def _on_show_detail(self):
            dlg = QTextEdit(self)
            dlg.setReadOnly(True)
            geo = QApplication.desktop().screenGeometry()
            dlg.setGeometry(geo.width() / 2, geo.height() / 2, geo.width() / 4, geo.height() / 4)
            dlg.setWindowTitle('Error Detail Information')
            dlg.setText(self.property('history'))
            dlg.setWindowFlags(Qt.Dialog)
            dlg.show()

        self.setProperty('theme_type', 'error')
        self.setReadOnly(True)
        _suffix_btn = buttons.BaseToolButton().image('delete_message').icon_only()
        _suffix_btn.clicked.connect(partial(_on_show_detail, self))
        self.set_suffix_widget(_suffix_btn)
        self.setPlaceholderText('Error information will be here ...')

        return self

    def tiny(self):
        """
        Sets line edit to tiny size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.tiny if widget_theme else theme.Theme.Sizes.TINY

        return self

    def small(self):
        """
        Sets line edit to small size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.small if widget_theme else theme.Theme.Sizes.SMALL

        return self

    def medium(self):
        """
        Sets line edit to medium size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.medium if widget_theme else theme.Theme.Sizes.MEDIUM

        return self

    def large(self):
        """
        Sets line edit to large size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.large if widget_theme else theme.Theme.Sizes.LARGE

        return self

    def huge(self):
        """
        Sets line edit to huge size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.huge if widget_theme else theme.Theme.Sizes.HUGE

        return self

    def password(self):
        """
        Sets line edit password mode
        """

        self.setEchoMode(QLineEdit.Password)

        return self

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_delay_text_changed(self):
        """
        Internal callback function that is called when delay timer is completed
        """

        self.delayTextChanged.emit(self.text())


class StyledLineEdit(QLineEdit, object):
    """
    Styled line edit that takes a different color if it's empty
    """

    def __init__(self, default='', off_color=(125, 125, 125), on_color=(255, 255, 255), parent=None):
        super(StyledLineEdit, self).__init__(parent=parent)

        self._value = ''
        self._default = ''
        self._off_color = off_color
        self._on_color = on_color

        self.set_default(default)
        self.textChanged.connect(self._on_change)

    def get_value(self):
        if self.text() == self._default:
            return ''
        else:
            return self._value

    def set_value(self, value):
        self._value = value

    value = property(get_value, set_value)

    def focusInEvent(self, event):
        if self.text() == self._default:
            self.setText('')
            self.setStyleSheet(self._get_on_style())

    def focusOutEvent(self, event):
        if self.text() == '':
            self.setText(self._default)
            self.setStyleSheet(self._get_off_style())

    def set_default(self, text):
        self.setText(text)
        self._default = text
        self.setStyleSheet(self._get_off_style())

    def _get_on_style(self):
        return 'QLineEdit{color:rgb(%s, %s, %s);}' % (self._on_color[0], self._on_color[1], self._on_color[2])

    def _get_off_style(self):
        return 'QLineEdit{color:rgb(%s, %s, %s);}' % (self._off_color[0], self._off_color[1], self._off_color[2])

    def _on_change(self, text):
        if text != self._default:
            self.setStyleSheet(self._get_on_style())
            self._value = text
        else:
            self.setStyleSheet(self._get_off_style())


class ClickLineEdit(QLineEdit, object):
    """
    Custom QLineEdit that becomes editable on click or double click
    """

    def __init__(self, text, single=False, double=False, pass_through_click=True):
        super(ClickLineEdit, self).__init__(text)

        self.setReadOnly(True)
        self._editing_style = self.styleSheet()
        self._default_style = "QLineEdit {border: 0;}"
        self.setStyleSheet(self._default_style)
        self.setContextMenuPolicy(Qt.NoContextMenu)

        if single:
            self.mousePressEvent = self.editEvent
        else:
            if pass_through_click:
                self.mousePressEvent = self._mouse_click_pass_through
        if double:
            self.mouseDoubleClickEvent = self.editEvent
        else:
            if pass_through_click:
                self.mousePressEvent = self._mouse_click_pass_through

        self.editingFinished.connect(self._on_edit_finished)

    def focusOutEvent(self, event):
        super(ClickLineEdit, self).focusOutEvent(event)
        self._on_edit_finished()

    def mousePressEvent(self, event):
        event.ignore()

    def mouseReleaseEvent(self, event):
        event.ignore()

    def editEvent(self, event):
        self.setStyleSheet(self._editing_style)
        self.selectAll()
        self.setReadOnly(False)
        self.setFocus()
        event.accept()

    def _mouse_click_pass_through(self, event):
        event.ignore()

    def _on_edit_finished(self):
        self.setReadOnly(True)
        self.setStyleSheet(self._default_style)
        self.deselect()


class BaseAttrLineEdit(QLineEdit, object):
    attr_type = None
    valueChanged = Signal()

    def __init__(self, parent=None):
        super(BaseAttrLineEdit, self).__init__(parent=parent)

        self.returnPressed.connect(self.update)
        self.editingFinished.connect(self.update)

    def get_value(self):
        return None

    value = property(get_value)


class FloatLineEdit(BaseAttrLineEdit, object):
    attr_type = 'float'
    valueChanged = Signal(float)

    def __init__(self, parent=None):
        super(FloatLineEdit, self).__init__(parent=parent)

    # region Override Properties
    def get_value(self):
        if not self.text():
            return 0.0
        return float(self.text())

    value = property(get_value)

    def setText(self, text):
        super(FloatLineEdit, self).setText('%.2f' % float(text))

    def update(self):
        if self.text():
            self.setText(self.text())
        super(FloatLineEdit, self).update()
        self.valueChanged.emit(float(self.text()))


class IntLineEdit(QLineEdit, object):
    attr_type = 'int'
    valueChanged = Signal(int)

    def __init__(self, parent=None):
        super(IntLineEdit, self).__init__(parent=parent)

    def get_value(self):
        if not self.text():
            return 0
        return int(self.text())

    value = property(get_value)

    def setText(self, text):
        super(IntLineEdit, self).setText('%s' % int(text))

    def update(self):
        if self.text():
            self.setText(self.text())
        super(IntLineEdit, self).update()
        self.valueChanged.emit(int(self.text()))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for custom combo boxes
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.qt.core import mixin, formatters, theme


@mixin.theme_mixin
@mixin.cursor_mixin
@mixin.property_mixin
class BaseComboBox(QComboBox, object):

    valueChanged = Signal(list)

    def __init__(self, parent=None):
        super(BaseComboBox, self).__init__(parent)

        self._root_menu = None
        self._display_formatter = formatters.display_formatter
        self._has_custom_view = False
        self._size = self.theme_default_size()

        self.setEditable(True)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)

        line_edit = self.lineEdit()
        line_edit.setReadOnly(True)
        line_edit.setTextMargins(4, 0, 4, 0)
        line_edit.setStyleSheet('background-color: transparent;')
        line_edit.installEventFilter(self)

        self.set_value('')
        self.set_placeholder('Please Select')

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    def _get_size(self):
        """
        Returns the button height size
        :return: float
        """

        return self._size

    def _set_size(self, value):
        """
        Sets button height size
        :param value: float
        """

        self._size = value
        self.lineEdit().setProperty('theme_size', value)
        self.style().polish(self)

    theme_size = Property(int, _get_size, _set_size)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def eventFilter(self, widget, event):
        """
        Overrides base eventFilter function
        :param widget:
        :param event:
        :return:
        """

        if widget is self.lineEdit():
            if event.type() == QEvent.MouseButtonPress:
                self.showPopup()
        return super(BaseComboBox, self).eventFilter(widget, event)

    def setView(self, *args, **kwargs):
        """
        Overrides base setView function
        :param args:
        :param kwargs:
        """

        self._has_custom_view = True
        super(BaseComboBox, self).setView(*args, **kwargs)

    def showPopup(self):
        """
        Overrides base showPopup function.
        If we have a custom menu, we make sure that we show it.
        """

        if self._has_custom_view or self._root_menu is None:
            super(BaseComboBox, self).showPopup()
        else:
            QComboBox.hidePopup(self)
            self._root_menu.popup(self.mapToGlobal(QPoint(0, self.height())))

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_formatter(self, fn):
        """
        Sets the formatter used by combobox
        :param fn:
        :return:
        """

        self._display_formatter = fn

    def set_placeholder(self, text):
        """
        Sets the placeholder text that appears when no item is selected
        :param text: str
        """

        self.lineEdit().setPlaceholderText(text)

    def set_value(self, value):
        """
        Sets combo box value
        :param value:
        """

        self.setProperty('value', value)

    def set_menu(self, menu):
        """
        Sets combo box custom menu
        :param menu: QMenu
        """

        self._root_menu = menu
        self._root_menu.valueChanged.connect(self.valueChanged)
        self._root_menu.valueChanged.connect(self.set_value)

    def tiny(self):
        """
        Sets spin box to tiny size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.tiny if widget_theme else theme.Theme.Sizes.TINY

        return self

    def small(self):
        """
        Sets spin box to small size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.small if widget_theme else theme.Theme.Sizes.SMALL

        return self

    def medium(self):
        """
        Sets spin box to medium size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.medium if widget_theme else theme.Theme.Sizes.MEDIUM

        return self

    def large(self):
        """
        Sets spin box to large size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.large if widget_theme else theme.Theme.Sizes.LARGE

        return self

    def huge(self):
        """
        Sets spin box to huge size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.huge if widget_theme else theme.Theme.Sizes.HUGE

        return self

    # =================================================================================================================
    # PROPERTY MIXIN SETTERS
    # =================================================================================================================

    def _set_value(self, value):
        self.lineEdit().setProperty('text', self._display_formatter(value))
        if self._root_menu:
            self._root_menu.set_value(value)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to create different types of lists
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *


class EditableList(QTreeWidget):
    """
    List list with editable list items
    """

    itemUpdated = Signal(str, str)
    editing = False

    class EditLine(QLineEdit):
        def __init__(self, *args):
            super(EditableList.EditLine, self).__init__(*args)

        def keyPressEvent(self, event):
            if event.key() == Qt.Key_Escape:
                self.deleteLater()
            elif event.key() in (Qt.Key_Return, Qt.Key_Enter):
                self.parent().parent().update_item()
            else:
                super(EditableList.EditLine, self).keyPressEvent(event)

    def __init__(self, parent=None):
        super(EditableList, self).__init__(parent=parent)
        self.itemClicked.connect(self._on_click)
        self.itemDoubleClicked.connect(self._on_double_click)

    def update_item(self):
        """
        If an item has been edited we update the list
        """

        i = self.editing
        w = self.itemWidget(i, 0)

        if w and not self.findItems(w.text(), Qt.MatchExactly, 0):
            self.itemUpdated.emit(i.text(0), w.text())
            i.setText(0, w.text())
            self.scrollToItem(i)
            self.editing = False
        if i:
            self.removeItemWidget(i, 0)

    def _on_click(self, item, col):
        if self.editing and item is not self.editing:
            self.update_item()

    def _on_double_click(self, item, col):
        edit = EditableList.EditLine(item.text(0), self)
        self.setItemWidget(item, 0, edit)
        edit.selectAll()
        self.editing = item


class WidgetItem(QFrame, object):

    closed = Signal(QWidget)
    deleted = Signal(QWidget)

    def __init__(self, item_height=150, item_width=300, height_offset=10, width_offset=10,
                 has_title=True, editable_title=True, is_closable=True, parent=None):
        super(WidgetItem, self).__init__(parent=parent)

        self._item_height = item_height
        self._item_width = item_width
        self._height_offset = height_offset
        self._width_offset = width_offset
        self._has_title = has_title
        self._editable_title = editable_title
        self._is_closable = is_closable
        self._animation = None

        self.setFrameStyle(QFrame.Panel | QFrame.Raised)

        self.custom_ui()

    def custom_ui(self):

        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(3, 1, 3, 3)
        self.layout().setSpacing(0)

        self.main_layout = QVBoxLayout()
        self.main_layout.setContentsMargins(2, 2, 2, 2)
        self.main_layout.setSpacing(5)
        main_widget = QWidget()

        if self._item_height:
            main_widget.setFixedHeight(self._item_height - self._height_offset)
        if self._item_width:
            main_widget.setFixedWidth(self._item_width - self._width_offset)

        main_widget.setLayout(self.main_layout)
        self.layout().addWidget(main_widget)

        # =====================================================

        # This layout is used to add custom widgets before the title of the node
        self.buttons_layout = QVBoxLayout()
        self.buttons_layout.setContentsMargins(2, 2, 2, 2)
        self.buttons_layout.setSpacing(2)

        title_layout = QHBoxLayout()
        title_layout.addLayout(self.buttons_layout)

        title_line = QLineEdit('Untitled')
        if not self._editable_title:
            title_line.setEnabled(False)

        if self._has_title:
            title_layout.addWidget(title_line)

        self._close_btn = QPushButton('X')
        self._close_btn.setFixedHeight(20)
        self._close_btn.setFixedWidth(20)
        self._close_btn.clicked.connect(self.close_widget)
        if self._is_closable:
            title_layout.addWidget(self._close_btn)

        self.main_layout.addLayout(title_layout)

    def hide_close_button(self, value=True):
        self._close_btn.setVisible(not(value))

    def close_widget(self):
        if self._is_closable:
            self.closed.emit(self)

    def delete_widget(self):
        self.deleted.emit(self)

    def _animate_expand(self, value):

        size_anim = QPropertyAnimation(self, 'geometry')
        geometry = self.geometry()
        width = geometry.width()
        x, y, _, _ = geometry.getCoords()
        size_start = QRect(x, y, width, int(not(value)) * 150)
        size_end = QRect(x, y, width, value * 150)
        size_anim.setStartValue(size_start)
        size_anim.setEndValue(size_end)
        size_anim.setDuration(300)
        size_anim_curve = QEasingCurve()
        if value:
            size_anim_curve.setType(QEasingCurve.InQuad)
        else:
            size_anim_curve.setType(QEasingCurve.OutQuad)
        size_anim.setEasingCurve(size_anim_curve)

        # =================================================== Animation Sequence

        self._animation = QSequentialAnimationGroup()
        self._animation.addAnimation(size_anim)
        size_anim.valueChanged.connect(self._force_resize)
        if not value:
            self._animation.finished.connect(self.delete_widget)
        self._animation.start(QAbstractAnimation.DeleteWhenStopped)

    def _force_resize(self, new_height):

        # Force widget item parent to reevaluate its size
        self.setFixedHeight(new_height.height())


class ItemWidgetsList(QListWidget, object):
    def __init__(self, parent=None):
        super(ItemWidgetsList, self).__init__(parent=parent)

        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setDefaultDropAction(Qt.MoveAction)
        self.setFocusPolicy(Qt.NoFocus)

    def addWidget(self, widget):

        item = QListWidgetItem(self)
        item.setSizeHint(widget.sizeHint())
        super(ItemWidgetsList, self).addItem(item)


class WidgetsList(QWidget, object):

    def __init__(self, parent=None):
        super(WidgetsList, self).__init__(parent=parent)

        self._widgets_list = list()

        self.main_layout = QVBoxLayout()
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self.main_layout.setAlignment(Qt.AlignTop)
        self.setLayout(self.main_layout)

        self.widgets_layout = QVBoxLayout()
        self.widgets_layout.setContentsMargins(0, 0, 0, 0)
        self.widgets_layout.setSpacing(0)
        self.main_layout.addLayout(self.widgets_layout)

        button_layout = QHBoxLayout()
        button_layout.setContentsMargins(5, 5, 5, 5)
        button_layout.setAlignment(Qt.AlignRight)
        self.main_layout.addLayout(button_layout)

        self.add_button = QPushButton('New...', parent=self)
        button_layout.addWidget(self.add_button)

    def add_widget(self, widget):
        self.widgets_layout.addWidget(widget)
        self._widgets_list.append(widget)

        # The first widget of the list cannot be deleted
        if len(self._widgets_list) <= 1:
            self._widgets_list[0].hide_close_button(True)

        # When the close button is pressed we delete the widget
        widget.closed.connect(self.remove_widget)

        widget.setFixedHeight(0)
        widget._animate_expand(True)

    def remove_widget(self, widget):

        # We use try/catch to avoid an error when closing/creating new widgets too quicly in the UI
        widget.deleted.connect(self.delete_widget)
        self._widgets_list.remove(widget)
        widget._animate_expand(False)

    def delete_widget(self, widget):
        self.widgets_layout.removeWidget(widget)
        widget._animation = None
        widget.deleteLater()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for DCC toolsets
"""

import os
from collections import OrderedDict

import tpDcc
from tpDcc import register
from tpDcc.core import plugin
from tpDcc.libs.python import python, decorators, folder, yamlio, color
from tpDcc.libs.qt.widgets import toolset

if python.is_python2():
    import pkgutil as loader
else:
    import importlib as loader


class ToolsetsPluginsManager(plugin.PluginsManager, object):

    INTERFACE = toolset.ToolsetWidget

    def __init__(self):
        super(ToolsetsPluginsManager, self).__init__()


class ToolsetsManager(object):
    def __init__(self,):
        super(ToolsetsManager, self).__init__()

        self._toolsets = dict()
        self._toolset_groups = dict()
        self._registered_paths = dict()

        self._manager = ToolsetsPluginsManager()

    @property
    def toolset_groups(self):
        return self._toolset_groups

    def register_path(self, package_name, path_to_register):
        """
        Register path to find toolsets from
        :return: str
        """

        if not path_to_register or not os.path.isdir(path_to_register):
            return None

        if package_name not in self._registered_paths:
            self._registered_paths[package_name] = list()

        if path_to_register not in self._registered_paths[package_name]:
            self._registered_paths[package_name].append(path_to_register)

    # ============================================================================================================
    # TOOLSETS
    # ============================================================================================================

    def load_registered_toolsets(self, package_name, tools_to_load, dev=True, do_reload=False, tools_manager=None):
        self._load_registered_paths_toolsets(package_name=package_name, do_reload=do_reload)
        self._load_package_toolsets(
            package_name=package_name, tools_to_load=tools_to_load, dev=dev, do_reload=do_reload)

        tools_mgr = tools_manager or tpDcc.ToolsMgr

        for pkg_name, toolset_data in self._manager.plugins.items():
            if package_name != pkg_name:
                continue
            if package_name not in self._toolsets:
                self._toolsets[package_name] = list()
            for tool_set in toolset_data.values():
                if tool_set.ID not in self._toolsets[package_name]:
                    toolset_config = tools_mgr().get_tool_config(tool_set.ID, package_name=package_name)
                    if not toolset_config:
                        tpDcc.logger.warning(
                            'No valid configuration file found for toolset: "{}" in package: "{}"'.format(
                                tool_set.ID, package_name))
                        continue
                    tool_set.CONFIG = toolset_config
                    self._toolsets[package_name].append({tool_set.ID: tool_set})

        return True

    def toolset(self, toolset_id, package_name=None, as_dict=False):
        """
        Returns toolset based on its ID
        :param toolset_id: str
        :param package_name: str
        :param as_dict: bool
        :return: ToolsetWidget
        """

        if not package_name:
            package_name = toolset_id.replace('.', '-').split('-')[0]

        package_toolsets = self._toolsets.get(package_name)
        if not package_toolsets:
            tpDcc.logger.warning('No toolsets found in package: {}!'.format(package_name))
            return None

        toolset_found = None
        for package_toolset in package_toolsets:
            if toolset_id in package_toolset:
                toolset_found = package_toolset
                break

        if not toolset_found:
            raise KeyError('"Toolset "{}" not found in package: "{}".'.format(toolset_id, package_name))

        if as_dict:
            return toolset_found
        else:
            return toolset_found[toolset_id]

    def toolset_ids(self, group_type, package_name=None):
        """
        Returns all toolsets in a given group type
        :param group_type: str
        :param package_name: str
        :return: list
        """

        if not self._toolset_groups:
            return None
        if package_name and package_name not in self._toolset_groups:
            return None

        toolsets_found = list()
        for pkg_name, toolset_groups in self._toolset_groups.items():
            for toolset_group in toolset_groups:
                if package_name and package_name != pkg_name:
                    continue
                if toolset_group['type'] == group_type:
                    toolsets_found.extend(toolset_group['toolsets'])

        return list(set(toolsets_found))

    def toolsets(self, group_type, package_name=None, as_dict=False):
        """
        Returns all toolsets available in a given group type
        :param group_type: str
        :param package_name: str
        :param as_dict: bool
        :return: list(ToolsetWidget)
        """

        toolset_ids = self.toolset_ids(group_type, package_name=package_name)
        return [self.toolset(toolset_id, package_name=package_name, as_dict=as_dict) for toolset_id in toolset_ids]

    def toolset_widgets(self, package_name=None, sort=True):
        """
        Returns a list with toolset widgets
        :param package_name: str
        :param sort: bool
        :return:
        """

        tolset_widgets = list()

        if package_name and package_name not in self._toolsets:
            return

        if package_name:
            tolset_widgets = self._toolsets[package_name].values()
        else:
            for package_name, toolsets in self._toolsets.items():
                for toolset in toolsets:
                    for toolset_id, toolset_widget in toolset.items():
                        tolset_widgets.append(toolset_widget)

        if sort:
            tolset_widgets.sort(key=lambda toolset: toolset.CONFIG.get('name'))

        return tolset_widgets

    def toolset_color(self, toolset_id, package_name=None):
        """
        Returns the color of the toolset and shifts it if necessary
        :param toolset_id: str
        :param package_name: str
        :return: tuple(float, float, float)
        """

        if not toolset_id:
            return 255, 255, 255

        if not package_name:
            package_name = toolset_id.replace('.', '-').split('-')[0]

        if self._toolset_groups and package_name in self._toolset_groups:
            for pkg_name, toolset_groups in self._toolset_groups.items():
                if package_name != pkg_name:
                    continue
                for toolset_group in toolset_groups:
                    if toolset_id in toolset_group['toolsets']:
                        index = toolset_group['toolsets'].index(toolset_id)
                        group_color = tuple(toolset_group['color'])
                        hue_shift = toolset_group['hue_shift'] * (index + 1)
                        return tuple(color.hue_shift(group_color, hue_shift))
        else:
            tpDcc.logger.warning(
                'ToolSet "{}" not found in any toolset group. Impossible to retrieve color!'.format(toolset_id))
            return 255, 255, 255

    def group_types(self):
        found_group_types = list()
        for pkg_name, toolset_groups in self._toolset_groups.items():
            for toolset_group in toolset_groups:
                found_group_types.append(toolset_group['name'])

        return found_group_types

    def group_type(self, group_name):
        """
        Returns type by group name
        :param group_name: str
        :return: str
        """

        for pkg_name, toolset_groups in self._toolset_groups.items():
            for toolset_group in toolset_groups:
                if toolset_group['name'] == group_name:
                    return toolset_group['type']

        return None

    def group_color(self, group_type, package_name=None):
        """
        Returns color by group type
        :param group_type: str
        :return: str
        """

        if not self._toolset_groups:
            return

        for pkg_name, toolset_groups in self._toolset_groups.items():
            if package_name and pkg_name != package_name:
                continue
            for toolset_group in toolset_groups:
                if toolset_group['type'] == group_type:
                    return toolset_group['color']

        return None

    def group_from_toolset(self, toolset_id, package_name=None):
        """
        Returns group of given toolset
        :param toolset_id: str
        :param package_name: str
        :return: str
        """

        if not self._toolset_groups:
            return

        for pkg_name, toolset_groups in self._toolset_groups.items():
            if package_name and pkg_name != package_name:
                continue
            for toolset_group in toolset_groups:
                for tool_set in toolset_group['toolsets']:
                    if tool_set == toolset_id:
                        return toolset_group['type']

        return None

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _load_registered_paths_toolsets(self, package_name, do_reload=False):
        """
        Loads all toolsets found in registered paths
        """

        if not self._registered_paths:
            return

        # Load toolsets data
        for pkg_name, registered_paths in self._registered_paths.items():
            if package_name != pkg_name:
                continue
            for registered_path in registered_paths:
                if not registered_path or not os.path.isdir(registered_path):
                    continue
                for pth in folder.get_files_with_extension(
                        toolset.ToolsetWidget.EXTENSION, registered_path, full_path=True, recursive=True):
                    try:
                        toolset_data = yamlio.read_file(pth, maintain_order=True)
                    except Exception:
                        tpDcc.logger.warning('Impossible to read toolset data from: "{}!'.format(pth))
                        continue
                    if pkg_name not in self._toolset_groups:
                        self._toolset_groups[pkg_name] = list()
                    toolset_type = toolset_data.get('type')
                    toolset_not_added = True
                    for _, toolset_groups in self._toolset_groups.items():
                        for toolset_group in toolset_groups:
                            if toolset_type == toolset_group['type']:
                                toolset_not_added = False
                                break
                        if not toolset_not_added:
                            break
                    if toolset_not_added:
                        self._toolset_groups[pkg_name].append(toolset_data)

        # Load toolsets widgets
        for pkg_name, registered_paths in self._registered_paths.items():
            if package_name != pkg_name:
                continue
            self._manager.register_paths(registered_paths, package_name=pkg_name, do_reload=do_reload)

    def _load_package_toolsets(self, package_name, tools_to_load, dev=True, do_reload=False):
        """
        Loads all toolsets available in given package
        :param package_name: str
        :param tools_to_load: lsit
        :param dev: bool
        """

        if not tools_to_load:
            return
        tools_to_load = python.force_list(tools_to_load)

        paths_to_register = OrderedDict()

        tools_path = '{}.tools.{}'
        tools_paths_to_load = list()
        for tool_name in tools_to_load:
            pkg_path = tools_path.format(package_name, tool_name)
            pkg_loader = loader.find_loader(pkg_path)
            if not pkg_loader:
                continue
            if tool_name not in paths_to_register:
                paths_to_register[tool_name] = list()

            if pkg_loader.filename not in tools_paths_to_load:
                tools_paths_to_load.append(pkg_loader.filename)

        # Find toolset widgets are located
        if tools_paths_to_load:
            self._manager.register_paths(tools_paths_to_load, package_name=package_name, do_reload=do_reload)


@decorators.Singleton
class ToolsetsManagerSingleton(ToolsetsManager, object):
    """
    Singleton class that holds preferences manager instance
    """

    def __init__(self):
        ToolsetsManager.__init__(self)


register.register_class('ToolsetsMgr', ToolsetsManagerSingleton)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utility functions related with images
"""

from __future__ import print_function, division, absolute_import

import os
import re
import base64
import logging
import traceback

try:
    import urllib2 as urllib
except ImportError:
    import urllib

from Qt.QtCore import *
from Qt.QtGui import *

from tpDcc.libs import qt
from tpDcc.libs.python import python, path as path_utils

LOGGER = logging.getLogger()

ImageFormats = {
    QImage.Format_Mono: 'L',                            # Mono
    QImage.Format_MonoLSB: 'L',                         # Mono LSB
    QImage.Format_Indexed8: 'P',                        # 8bit Color Map
    QImage.Format_RGB32: 'RGB',                         # RGB32
    QImage.Format_ARGB32: 'RGBA',                       # RGBA32
    QImage.Format_ARGB32_Premultiplied: 'RGBA',         # RGBA_Premultiplied
    QImage.Format_RGB16: 'RGB',                         # RGB16
    QImage.Format_ARGB8565_Premultiplied: 'RGBA',       # RGBA24 Premultiplied (5-6-5-8)
    QImage.Format_RGB666: 'RGB',                        # RGB24 (6-6-6)
    QImage.Format_ARGB6666_Premultiplied: 'RGBA',       # RGBA24 Premultiplied (6-6-6-6)
    QImage.Format_RGB555: 'RGB',                        # RGB16 (5-5-5)
    QImage.Format_ARGB8555_Premultiplied: 'RGBA',       # RGB24 Premultiplied (5-5-5-8)
    QImage.Format_RGB888: 'RGB',                        # RGB (8-8-8)
    QImage.Format_RGB444: 'RGB',                        # RGB (4-4-4)
    QImage.Format_ARGB4444_Premultiplied: 'RGB'         # RGBA16 Premultiplied (4-4-4-4)
}


def mode_to_qformat(mode):
    """
    Returns specific image mode taking into account given mode string
    :param mode: str
    :return: QImage.Format
    """

    if mode == 'RGBA':
        return QImage.Format_ARGB32_Premultiplied
    else:
        return QImage.Format_RGB32


def create_image(width, height, mode):
    """
    Creates a new QImage with the given resolution and mode
    :param width: int
    :param height: int
    :param mode: str
    :return: QImage
    """

    mode = mode_to_qformat(mode)
    if width <= 0 or height <= 0:
        raise Exception('Resolution for new image is negative or zero (X: {}, Y: {})'.format(width, height))
    new_image = QImage(width, height, mode)

    return new_image


def create_empty_image(output=None, resolution_x=1920, resolution_y=1080, background_color=None):
    """
    Creates an empty image and stores it in the given path
    :param output: str
    :param resolution_x: int
    :param resolution_y: int
    :param background_color: list(int, int, int)
    :return: str or QImage
    """

    if background_color is None:
        background_color = [0, 0, 0]

    pixmap = QPixmap(resolution_x, resolution_y)
    pixmap.fill(QColor(*background_color))
    if output:
        output_path = path_utils.clean_path(output)
        output_dir = os.path.dirname(output_path)
        if os.access(output_dir, os.W_OK):
            pixmap.save(output_path)
            return output_path

    return QImage(pixmap)


def get_image_width(image_path):
    """
    Returns the width of the image stored in given file path
    :param image_path: str
    :return: float
    """

    if not image_path or not os.path.isfile(image_path):
        return 0

    image = QImage(image_path)
    if not image or image.isNull():
        return 0

    return image.width()


def get_image_height(image_path):
    """
    Returns the height of the image stored in given file path
    :param image_path: str
    :return: float
    """

    if not image_path or not os.path.isfile(image_path):
        return 0

    image = QImage(image_path)
    if not image or image.isNull():
        return 0

    return image.height()


def paint_background(image, background_color=None):
    """
    Paints a background of the given color into the given QImage
    :param image: QImage
    :param background_color: tuple(r, g, b)
    :return: QImage, image with painted background
    """

    if not background_color:
        background_color = [0, 0, 0]

    if type(background_color) not in (list, tuple) or len(background_color) != 3:
        LOGGER.warning(
            'background_color argument is not valid ({}) using black color instead!'.format(background_color))
        background_color = [0, 0, 0]
    background_color = python.force_list(background_color)

    for i in range(len(background_color)):
        if background_color[i] < 0 or background_color[i] > 255:
            LOGGER.warning(
                'Background color channel ({})({}) out of limit (0, 255). Fixing to fit range ...'.format(
                    i, background_color[i]))
            background_color[i] = min(max(background_color[i], 0), 255)

    image.fill(QColor(*background_color).rgb())

    return image


def fit_image_in_resolution(image, width, height):
    """
    Resizes given image to fit in given width and height.
    :param image: QImage
    :param width: int
    :param height: int
    :return: QImage
    """

    resize_ratio = min((float(width) / float(image.width())), float(height) / float(image.height()))
    new_width = int(round(float(image.width()) * resize_ratio))
    new_height = int(round(float(image.height()) * resize_ratio))

    return image.scaled(new_width, new_height, Qt.KeepAspectRatio)


def overlay_image(front_image, back_image, x, y):
    """
    Overlays front image on top of given background image
    :param front_image: QImage
    :param back_image: QImage
    :param x: int
    :param y: int
    """

    painter = QPainter(back_image)
    painter.setRenderHint(QPainter.Antialiasing)
    painter.drawImage(x, y, front_image)
    painter.end()


def download_image(image_url, target_image_path):
    """
    Downloads image from URL and stores in given target path in disk
    :param image_url: str
    :param target_image_path: str
    """

    if not os.path.isfile(target_image_path):
        with open(target_image_path, 'wb') as temp_out_image:
            try:
                file_image_url = urllib.urlopen(image_url)
            except urllib.HTTPError as exc:
                raise Exception('HTTPError: {} | {}'.format(image_url, exc.code))
            except urllib.URLError as exc:
                raise Exception('URLError: {} | {}'.format(image_url, exc))
            except Exception as exc:
                raise Exception('Generic exception: {} | {}'.format(image_url, traceback.format_exc()))

            temp_out_image.write(file_image_url.read())
            file_image_url.close()


class ImageWorker(QRunnable, object):
    """
    Class that loads an image in a thread
    """

    class ImageWorkerSignals(QObject, object):
        triggered = Signal(object)

    def __init__(self, *args):
        super(ImageWorker, self).__init__(*args)

        self._path = None
        self.signals = ImageWorker.ImageWorkerSignals()

    def set_path(self, path):
        """
        Set the image path to be loaded
        :param path: str
        """

        self._path = path

    def run(self):
        """
        Overrides base QRunnable run function
        This is the starting point for the thread
        """

        try:
            if self._path:
                image = QImage(str(self._path))
                self.signals.triggered.emit(image)
        except Exception as e:
            qt.logger.error('Cannot load thumbnail image!')


class ImageSequence(QObject, object):

    DEFAULT_FPS = 24

    frameChanged = Signal(int)

    def __init__(self, path, *args):
        super(ImageSequence, self).__init__(*args)

        self._fps = self.DEFAULT_FPS
        self._timer = None
        self._frame = 0
        self._frames = list()
        self._dirname = None
        self._paused = False

        if path:
            self.set_dirname(path)

    def set_path(self, path):
        """
        Set as singal frame image sequence
        :param path: str
        """

        self._frame = 0
        self._frames = [path]

    def dirname(self):
        """
        Return the location to the image sequence in disk
        :return: str
        """

        return self._dirname

    def set_dirname(self, dirname):
        """
        Set the location where image sequence files are located
        :param dirname: str
        """

        def natural_sort_items(items):
            """
            Sort the given list in the expected way
            :param items: list(str)
            """

            def _convert(text):
                return int(text) if text.isdigit() else text

            def _alphanum_key(key):
                return [_convert(c) for c in re.split('([0-9]+)', key)]

            items.sort(key=_alphanum_key)

        self._dirname = dirname
        if os.path.isdir(dirname):
            self._frames = [dirname + '/' + filename for filename in os.listdir(dirname)]
            natural_sort_items(self._frames)

    def start(self):
        """
        Starts the image sequence
        """

        self.reset()
        if self._timer:
            self._timer.start(1000.0 / self._fps)

    def pause(self):
        """
        Pause image sequence
        """

        self._paused = True
        if self._timer:
            self._timer.stop()

    def resume(self):
        """
        Play image sequence after Pause
        """

        if self._paused:
            self._paused = False
            if self._timer:
                self._timer.start()

    def stop(self):
        """
        Stop the image sequence
        """

        if self._timer:
            self._timer.stop()

    def reset(self):
        """
        Stop and reset the current frame to 0
        """

        if not self._timer:
            self._timer = QTimer(self.parent())
            self._timer.setSingleShot(False)
            self._timer.timeout.connect(self._on_frame_changed)

        if not self._paused:
            self._frame = 0
        self._timer.stop()

    def frames(self):
        """
        Return all the filenames in the image sequence
        :return: list(str)
        """

        return self._frames

    def percent(self):
        """
        Return the current frame position as a percentage
        :return: float
        """

        if len(self._frames) == self._frame + 1:
            _percent = 1
        else:
            _percent = float((len(self._frames) + self._frame)) / len(self._frames) - 1

        return _percent

    def frame_count(self):
        """
        Returns the number of frames
        :return: int
        """

        return len(self._frames)

    def current_frame_number(self):
        """
        Returns the current frame
        :return: int
        """

        return self._frame

    def current_filename(self):
        """
        Returns the current file name
        :return: str
        """

        try:
            return self._frames[self.current_frame_number()]
        except IndexError:
            pass

    def current_icon(self):
        """
        Returns the current frames as QIcon
        :return: QIcon
        """

        return QIcon(self.current_filename())

    def current_pixmap(self):
        """
        Returns the current frame as QPixmap
        :return: QPixmap
        """

        return QPixmap(self.current_filename())

    def jump_to_frame(self, frame):
        """
        Set the current frame
        :param frame: int
        """

        if frame >= self.frame_count():
            frame = 0
        self._frame = frame
        self.frameChanged.emit(frame)

    def _on_frame_changed(self):
        """
        Internal callback function that is called when the current frame changes
        """

        if not self._frames:
            return

        frame = self._frame
        frame += 1
        self.jump_to_frame(frame)


def image_to_base64(image_path):
    """
    Converts image file to base64
    :param image_path: str
    :return: str
    """

    if os.path.isfile(image_path):
        with open(image_path, 'rb') as image_file:
            return base64.b64encode(image_file.read())


def base64_to_image(base64_string, image_format='PNG'):
    """
    Converts base64 to QImage
    :param base64_string: str
    :param image_format: str
    :return: QImage
    """

    if isinstance(base64_string, basestring):
        ba = QByteArray.fromBase64(base64_string)
        image = QImage.fromData(ba, image_format)
        return image


def base64_to_bitmap(base64_string, bitmap_format='PNG'):
    """
    Converts base64 to QBitmap
    :param base64_string: str
    :param image_format: str
    :return: QBitmap
    """

    if isinstance(base64_string, basestring):
        image = base64_to_image(base64_string, bitmap_format)
        if image is not None:
            bitmap = QBitmap.fromImage(image)
            return bitmap


def base64_to_icon(base64_string, icon_format='PNG'):
    """
    Converts base64 to QIcon
    :param base64_string: str
    :param icon_format: str
    :return: QIcon
    """

    if isinstance(base64_string, basestring):
        bitmap = base64_to_bitmap(base64_string, icon_format)
        if bitmap is not None:
            icon = QIcon(bitmap)
            return icon



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains toast widget implementation
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc as tp
from tpDcc.libs.qt.core import base, mixin, animation, theme
from tpDcc.libs.qt.widgets import label, avatar, loading


@mixin.theme_mixin
class BaseToast(base.BaseWidget, object):

    class ToastTypes(object):
        INFO = 'info'
        SUCCESS = 'success'
        WARNING = 'warning'
        ERROR = 'error'
        LOADING = 'loading'

    DEFAULT_CONFIG = {'duration': 2}

    toastClosed = Signal()

    def __init__(self, text, duration=None, toast_type=None, parent=None):
        self._text = text
        self._duration = duration
        self._toast_type = toast_type
        self._parent = parent
        super(BaseToast, self).__init__(parent=parent)

    def get_main_layout(self):
        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)

        return main_layout

    def ui(self):
        super(BaseToast, self).ui()

        self.setWindowFlags(Qt.FramelessWindowHint | Qt.Dialog | Qt.WA_TranslucentBackground | Qt.WA_DeleteOnClose)
        self.setAttribute(Qt.WA_StyledBackground)
        self.setFixedSize(QSize(120, 120))

        icon_layout = QHBoxLayout()
        icon_layout.addStretch()

        widget_theme = self.theme()

        if self._toast_type == self.ToastTypes.LOADING:
            icon_layout.addWidget(loading.CircleLoading(size=widget_theme.huge, color=widget_theme.text_color_inverse))
        else:
            icon_label = avatar.Avatar()
            icon_label.theme_size = 60
            icon_label.image = tp.ResourcesMgr().pixmap(
                self._toast_type or self.ToastTypes.INFO, color=widget_theme.text_color_inverse)
            icon_layout.addWidget(icon_label)
        icon_layout.addStretch()

        content_label = label.BaseLabel()
        content_label.setText(self._text or '')
        content_label.setAlignment(Qt.AlignCenter)

        self.main_layout.addStretch()
        self.main_layout.addLayout(icon_layout)
        self.main_layout.addSpacing(10)
        self.main_layout.addWidget(content_label)
        self.main_layout.addStretch()

        close_timer = QTimer(self)
        close_timer.setSingleShot(True)
        close_timer.timeout.connect(self.close)
        close_timer.timeout.connect(self.toastClosed.emit)
        close_timer.setInterval((self._duration or self.DEFAULT_CONFIG.get('duration', 2)) * 1000)

        anim_timer = QTimer(self)
        anim_timer.timeout.connect(self._fade_out)
        anim_timer.setInterval((self._duration or self.DEFAULT_CONFIG.get('duration', 2)) * 1000 - 300)

        self._opacity_anim = QPropertyAnimation()
        self._opacity_anim.setTargetObject(self)
        self._opacity_anim.setDuration(300)
        self._opacity_anim.setEasingCurve(QEasingCurve.OutCubic)
        self._opacity_anim.setPropertyName('windowOpacity')
        self._opacity_anim.setStartValue(0.0)
        self._opacity_anim.setEndValue(0.9)

        close_timer.start()
        anim_timer.start()

        self._get_center_position(self._parent)
        self._fade_in()

    @classmethod
    def info(cls, text, parent, duration=None):
        inst = cls(text, duration=duration, toast_type=cls.ToastTypes.INFO, parent=parent)
        inst.show()

        return inst

    @classmethod
    def success(cls, text, parent, duration=None):
        inst = cls(text, duration=duration, toast_type=cls.ToastTypes.SUCCESS, parent=parent)
        inst.show()

        return inst

    @classmethod
    def warning(cls, text, parent, duration=None):
        inst = cls(text, duration=duration, toast_type=cls.ToastTypes.WARNING, parent=parent)
        inst.show()

        return inst

    @classmethod
    def error(cls, text, parent, duration=None):
        inst = cls(text, duration=duration, toast_type=cls.ToastTypes.ERROR, parent=parent)
        inst.show()

        return inst

    @classmethod
    def loading(cls, text, parent, duration=None):
        inst = cls(text, duration=duration, toast_type=cls.ToastTypes.LOADING, parent=parent)
        inst.show()

        return inst

    @classmethod
    def config(cls, duration):
        if duration is not None:
            cls.DEFAULT_CONFIG['duration'] = duration

    def _fade_out(self):
        self._opacity_anim.setDirection(QAbstractAnimation.Backward)
        self._opacity_anim.start()

    def _fade_in(self):
        self._opacity_anim.start()

    def _get_center_position(self, parent):
        parent_parent = parent.parent()
        dcc_win = tp.Dcc.get_main_window()
        if dcc_win:
            dcc_window = parent_parent == dcc_win or parent_parent.objectName() == dcc_win.objectName()
        else:
            dcc_window = None
        parent_geo = parent.geometry()
        pos = parent_geo.topLeft() if dcc_window else parent.mapToGlobal(parent_geo.topLeft())
        offset = 0
        for child in parent.children():
            if isinstance(child, BaseToast) and child.isVisible():
                offset = max(offset, child.y())
        target_x = pos.x() + parent_geo.width() / 2 - self.width() / 2
        target_y = pos.y() + parent_geo.height() / 2 - self.height() / 2
        self.setProperty('pos', QPoint(target_x, target_y))


class ToastWidget(QLabel, object):
    """
    Toast widget used to show quick messages to user
    """

    DEFAULT_DURATION = 500
    DEFAULT_PADDING = 30

    def __init__(self, *args):
        super(ToastWidget, self).__init__(*args)

        self._timer = QTimer(self)
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self._on_fade_out)

        self._duration = self.DEFAULT_DURATION

        self.setMouseTracking(True)
        self.setAlignment(Qt.AlignCenter)
        self.setAttribute(Qt.WA_TransparentForMouseEvents)

        if self.parent():
            self.parent().installEventFilter(self)

    # def eventFilter(self, obj, event):
    #     """
    #     Overrides base QLabel eventFilter function
    #     Updates the geometry when the parent widget changes size
    #     :param obj: QWidget
    #     :param event: QEvent
    #     """
    #
    #     if event.type() == QEvent.Resize:
    #         self.updateGeometry()
    #     return super(ToastWidget, self).eventFilter(obj, event)

    def updateGeometry(self):
        """
        Overrides base QLabel updateGeometry function
        Updates and aligns the geometry to the parent widget
        """

        padding = self.DEFAULT_PADDING
        widget = self.parent()

        width = self.text_width() + padding
        height = self.text_height() + padding
        x = widget.width() * 0.5 - width * 0.5
        y = (widget.height() - height) / 1.2

        self.setGeometry(x, y, width, height)

    def setText(self, *args, **kwargs):
        """
        Overrides base QLabel setText function
        Updates the size depending on the text width
        :param text: str
        """

        super(ToastWidget, self).setText(*args, **kwargs)
        self.updateGeometry()

    def show(self):
        """
        Overrides base QLabel show function
        Starts the timer to hide the toast
        """

        duration = self.duration()
        self._timer.stop()
        self._timer.start(duration)
        if not self.isVisible():
            animation.fade_in_widget(self, duration=0)
            super(ToastWidget, self).show()

    def duration(self):
        """
        Returns duration
        :return: int
        """

        return self._duration

    def set_duration(self, duration):
        """
        Sets how long to show the toast (in milliseconds)
        :param duration: int
        """

        self._duration = duration

    def text_rect(self):
        """
        Returns the bounding box rect for the text
        :return: QRect
        """

        text = self.text()
        font = self.font()
        metrics = QFontMetricsF(font)

        return metrics.boundingRect(text)

    def text_width(self):
        """
        Returns the width of the text
        :return: int
        """

        text_width = self.text_rect().width()
        return max(0, text_width)

    def text_height(self):
        """
        Returns the height of the text
        :return: int
        """

        text_height = self.text_rect().height()
        return max(0, text_height)

    def _on_fade_out(self, duration=250):
        """
        Internal callback function that fades out the toast message
        :param duration: int
        """

        animation.fade_out_widget(self, duration=duration, on_finished=self.hide)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains consts exception used by libraries
"""

from __future__ import print_function, division, absolute_import

import six


class PathError(IOError, object):
    """
    Exception that supports unicode escape characters
    """

    def __init__(self, msg):
        msg = six.u(msg)
        super(PathError, self).__init__(msg)
        self._msg = msg

    def __unicode__(self):
        """
        Returns the decoded message using unicode_escape
        :return: str
        """

        msg = six.u(self._msg).decode('unicode_escape')
        return msg


class MovePathError(PathError):
    """
    Error related with path moving
    """

    pass


class RenamePathError(PathError):
    """
    Error related with path renaming
    """

    pass


class ItemError(Exception):
    pass


class ItemSaveError(Exception):
    pass


class ItemLoadError(Exception):
    pass


class DccUtilsError(Exception):
    pass


class ObjectsError(DccUtilsError):
    pass


class SelectionError(DccUtilsError):
    pass


class MoreThanOneObjectFoundError(DccUtilsError):
    pass


class ModelPanelNotInFocusError(DccUtilsError):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to create attribute editors
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs.python import strings as string_utils
from tpDcc.libs.qt.widgets import lineedit, directory
from tpDcc.libs.qt.core import qtutils, base, color


class AttributeEditor(base.BaseWidget, object):
    def __init__(self, name='AttributeEditor', label='Attribute Editor', parent=None):
        self._label = label
        self._objects = list()
        super(AttributeEditor, self).__init__(parent=parent)
        self.setObjectName(name)

    def get_objects(self):
        return self._objects

    objects = property(get_objects)

    def ui(self):
        super(AttributeEditor, self).ui()

        self._main_group = QGroupBox()
        self._main_group.setProperty('class', 'attr_main_group')
        self._main_group.setFlat(False)
        self._main_group.setTitle('')

        self._main_group_layout = self.get_attributes_layout()
        self._main_group.setLayout(self._main_group_layout)

        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setStyleSheet('QScrollArea { background-color: rgb(57,57,57);}')
        scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        scroll_area.setWidget(self._main_group)

        self.main_layout.addWidget(scroll_area)

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_create_context_menu)

    def get_attributes_layout(self):
        """
        Returns layout used for the attributes
        :return: QLayout
        """

        attributes_layout = QHBoxLayout()
        attributes_layout.setContentsMargins(5, 5, 5, 5)
        attributes_layout.setAlignment(Qt.AlignTop)

        return attributes_layout

    def set_title(self, new_title):
        self._main_group.setTitle(new_title)

    def clear_layout(self, reset_title=True):
        """
        Clears main layout and resetes editor title
        """

        qtutils.clear_layout(self._main_group_layout)
        if reset_title:
            self._main_group.setTitle('')

    def _build_layout(self):
        pass

    def _on_create_context_menu(self):
        pass


class BaseEditor(QWidget, object):
    attr_type = 'None'
    valueChanged = Signal(object)

    def __init__(self, parent=None, **kwargs):
        super(BaseEditor, self).__init__(parent=parent)

        self._default_value = 0.0
        self._current_value = None

        self.main_layout = QHBoxLayout(self)
        self.main_layout.setContentsMargins(1, 1, 1, 1)
        self.main_layout.setSpacing(3)

    def get_default_value(self):
        return self._default_value

    def set_default_value(self, value):
        if value != self._default_value:
            self._default_value = value

    def get_value(self):
        return None

    def set_value(self, value):
        pass

    default_value = property(get_default_value, set_default_value)
    value = property(get_value)

    def OnValueUpdated(self):
        if self.value != self._current_value:
            self._current_value = self.value
            self.valueChanged.emit(self)

    def initialize_editor(self):
        pass


class FloatEditor(BaseEditor, object):

    attr_type = 'float'

    def __init__(self, parent=None, **kwargs):
        super(FloatEditor, self).__init__(parent=parent, **kwargs)

        self._default_value = 0.0

        self.value_line = lineedit.FloatLineEdit(self)
        self.main_layout.addWidget(self.value_line)

    def get_value(self):
        return self.value_line.value

    value = property(get_value)

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]
                if node_values[0] is None:
                    editor_value = self.default_value

                self.value_line.blockSignals(True)
                self.value_line.setText(str(editor_value))
                self.value_line.blockSignals(False)
                self.value_line.valueChanged.connect(self.OnValueUpdated)

    def set_connected(self, conn):
        if conn != self._connection:
            self._connection = conn
            self.value_line.setText(conn)
            self.value_line.setEnabled(False)
            self.value_line.setProperty('class', 'Connected')


class Float2Editor(BaseEditor, object):

    attr_type = 'float'

    def __init__(self, parent=None, **kwargs):
        super(Float2Editor, self).__init__(parent=parent, **kwargs)

        self._default_value = (0.0, 0.0)

        self.value1_line = lineedit.FloatLineEdit(self)
        self.value2_line = lineedit.FloatLineEdit(self)
        self.main_layout.addWidget(self.value1_line)
        self.main_layout.addWidget(self.value2_line)

    def get_value(self):
        return self.value1_line.value, self.value2_line.value

    value = property(get_value)

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]

                self.value1_line.blockSignals(True)
                self.value2_line.blockSignals(True)

                if type(editor_value) in [QPoint, QPointF]:
                    self.value1_line.setText(str(editor_value.x()))
                    self.value2_line.setText(str(editor_value.y()))
                elif type(editor_value) in [list, tuple]:
                    self.value1_line.setText(str(editor_value[0]))
                    self.value2_line.setText(str(editor_value[1]))
                self.value1_line.blockSignals(False)
                self.value2_line.blockSignals(False)
                self.value1_line.valueChanged.connect(self.OnValueUpdated)
                self.value2_line.valueChanged.connect(self.OnValueUpdated)

    def set_connected(self, conn):
        if conn != self._connection:
            self._conection = conn
            self.value1_line.setText(conn)
            self.value1_line.setEnabled(False)
            self.value2_line.setEnabled(False)
            self.value1_line.setProperty('class', 'Connected')
            self.value2_line.setProperty('class', 'Connected')


class Float3Editor(BaseEditor, object):

    attr_type = 'float'

    def __init__(self, parent=None, **kwargs):
        super(Float3Editor, self).__init__(parent=parent, **kwargs)

        self._default_value = (0.0, 0.0, 0.0)

        self.value1_line = lineedit.FloatLineEdit(self)
        self.value2_line = lineedit.FloatLineEdit(self)
        self.value3_line = lineedit.FloatLineEdit(self)
        self.main_layout.addWidget(self.value1_line)
        self.main_layout.addWidget(self.value2_line)
        self.main_layout.addWidget(self.value3_line)

    def get_value(self):
        return self.value1_line.value(), self.value2_line.value(), self.value3_line.value()

    value = property(get_value)

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]

                self.value1_line.blockSignals(True)
                self.value2_line.blockSignals(True)
                self.value3_line.blockSignals(True)
                self.value1_line.setText(str(editor_value[0]))
                self.value2_line.setText(str(editor_value[1]))
                self.value3_line.setText(str(editor_value[2]))
                self.value1_line.blockSignals(False)
                self.value2line.blockSignals(False)
                self.value3line.blockSignals(False)
                self.value1_line.valueChanged.connect(self.OnValueUpdated)
                self.value2_line.valueChanged.connect(self.OnValueUpdated)
                self.value3_line.valueChanged.connect(self.OnValueUpdated)

    def set_connected(self, conn):
        if conn != self._connection:
            self._conection = conn
            self.value1_line.setText(conn)
            self.value1_line.setEnabled(False)
            self.value2_line.setEnabled(False)
            self.value3_line.setEnabled(False)
            self.value1_line.setProperty('class', 'Connected')
            self.value2_line.setProperty('class', 'Connected')
            self.value3_line.setProperty('class', 'Connected')


class StringEditor(BaseEditor, object):

    attr_type = 'str'

    def __init__(self, parent=None, **kwargs):
        super(StringEditor, self).__init__(parent=parent, **kwargs)

        self._default_value = ""
        self._clean_value = kwargs.get('clean', True)

        self.value_line = QLineEdit(self)
        regexp = QRegExp('^([a-zA-Z0-9_]+)')
        self.main_layout.addWidget(self.value_line)

        self.value_line.textEdited.connect(self._validate_text)
        self.value_line.editingFinished.connect(self.OnValueUpdated)
        self.value_line.returnPressed.connect(self.OnValueUpdated)

    def get_value(self):
        return str(self.value_line.text())

    value = property(get_value)

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]

                self.value_line.blockSignals(True)
                self.value_line.setText(str(editor_value))
                self.value_line.blockSignals(False)

    def set_connected(self, conn):
        if conn != self._connection:
            self._connection = conn
            self.value_line.setText(conn)
            self.value_line.setEnabled(False)
            self.value_line.setProperty('class', 'Connected')

    def _validate_text(self, text):
        """
        Validates the given value and update the current text
        :param text: str, text to validate
        """

        current_text = self.value
        if self._clean_value:
            cursor_pos = self.value_line.cursorPosition()
            cleaned = string_utils.clean_string(text=text)
            self.value_line.blockSignals(True)
            self.value_line.setText(cleaned)
            self.value_line.blockSignals(False)
            self.value_line.setCursorPosition(cursor_pos)


class BoolEditor(BaseEditor, object):

    attr_type = 'bool'

    def __init__(self, parent=None, **kwargs):
        super(BoolEditor, self).__init__(parent=parent, **kwargs)

        self._default_value = False

        self.cbx = QCheckBox(self)
        self.main_layout.addWidget(self.cbx)

        self.cbx.toggled.connect(self.OnValueUpdated)

    def get_value(self):
        return self.cbx.isChecked()

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]

                self.cbx.blockSignals(True)
                self.cbx.setChecked(editor_value)
                self.cbx.blockSignals(False)

    def set_connected(self, conn):
        if conn != self._connection:
            self._connection = conn
            self.value_line.setText(conn)
            self.cbx.setEnabled(False)
            self.value_line.setProperty('class', 'Connected')


class ColorPicker(BaseEditor, object):

    attr_type = 'color'

    def __init__(self, parent=None, **kwargs):
        super(ColorPicker, self).__init__(parent=parent, **kwargs)

        self._default_value = (125, 125, 125)
        self.attr = None

        self.normalized = kwargs.get('normalized', False)
        self.min = kwargs.get('min', 0)
        self.max = kwargs.get('max', 99)
        self.color = kwargs.get('color', QColor(1.0, 1.0, 1.0))
        self.mult = kwargs.get('mult', 0.1)

        self.color_swatch = color.ColorSwatch(parent=self, color=self.color, normalized=self.normalized)
        self.color_swatch.setMaximumSize(QSize(75, 20))
        self.color_swatch.setMinimumSize(QSize(75, 20))
        self.color_swatch.set_color(color=self.color)
        self.main_layout.addWidget(self.color_swatch)

        self.slider = QSlider(self)
        self.slider.setOrientation(Qt.Horizontal)
        self.slider.setValue(self.max)
        self.main_layout.addWidget(self.slider)

        self.set_max(self.max)
        self.set_min(self.min)

        self.slider.valueChanged.connect(self.OnSliderChanged)
        self.slider.sliderReleased.connect(self.OnSliderReleased)
        self.color_swatch.clicked.connect(self.OnColorPicked)

    # region Properties
    def get_rgb(self):
        return self.color_swatch.qcolor.getRgb()[0:3]

    def get_rgbF(self):
        return self.color_swatch.qcolor.getRgbF()[0:3]

    def get_hsv(self):
        return self.color_swatch.qcolor.getHsv()[0:3]

    def get_hsvF(self):
        return self.color_swatch.qcolor.getHsvF()[0:3]

    rgb = property(get_rgb)
    rgbF = property(get_rgbF)
    hsv = property(get_hsv)
    gsvF = property(get_hsvF)

    def OnSliderChanged(self):
        slider_value = float(self.slider.value())
        if not self._current_value:
            tp.logger.debug(
                'Caching color: (%d, %d, %d)' % (
                    self.color_swatch.color[0], self.color_swatch.color[1], self.color_swatch.color[2]))
            self._current_value = self.color_swatch.color
        current_color = QColor(*self._current_value)
        darker = 200 - slider_value
        new_color = current_color.darker(darker)
        self.color_swatch.qcolor = new_color
        self.color_swatch._update()

    def OnSliderReleased(self):
        """
        Updates items color when the slider handle is released
        """

        color = self.color_swatch.color
        self.color_swatch._update()
        self.OnValueUpdated()

    def OnColorPicked(self):
        """
        Event to call color picker
        """

        dialog = QColorDialog(self.color_swatch.qcolor, self)
        if dialog.exec_():
            self.color_swatch.setPalette(QPalette(dialog.currentColor()))
            self.color_swatch.qcolor = dialog.currentColor()
            self.color_swatch._update()
            self.OnValueUpdated()

    def get_value(self):
        return self.color_swatch.color

    def initialize_editor(self):
        editor_value = self.default_value

        node_values = self.values
        if node_values:
            if len(node_values) > 1:
                pass
            elif len(node_values) == 1:
                if node_values[0]:
                    editor_value = node_values[0]

                self.color_swatch.set_color(editor_value)

    def set_connected(self, conn):
        if conn != self._connection:
            self._connection = conn
            self.value_line.setText(conn)
            self.color_swatch.setEnabled(False)

    def sizeHint(self):
        return QSize(350, 27)

    def set_attr(self, value):
        self.attr = value
        return self.attr

    def get_attr(self):
        return self.attr

    def set_min(self, value):
        """
        Sets the slider minimum value
        :param value: int, slider minimum value
        """

        self.min = value
        self.slider.setMinimum(value)

    def set_max(self, value):
        """
        Sets the silder maximum value
        :param value: int, slider maximum value
        """

        self.max = value
        self.slider.setMaximum(value)

    def get_qcolor(self):
        return self.color_swatch.qcolor

    def set_color(self, value):
        return self.color_swatch.set_color(color=value)

    def _update(self):
        return self.color_swatch._update()


class FileEditor(BaseEditor, object):

    attr_type = 'file'

    def __init__(self, parent=None, **kwargs):
        super(FileEditor, self).__init__(parent=parent, **kwargs)

        line_widget = directory.SelectFile()
        self.main_layout.addWidget(line_widget)


# ===============================================================================

WIDGET_MAPPER = dict(
    float=[FloatEditor, 0.0],
    float2=[Float2Editor, [0.0, 0.0]],
    float3=[Float3Editor, [0.0, 0.0, 0.0]],
    bool=[BoolEditor, False],
    str=[StringEditor, ""],
    string=[StringEditor, ""],
    file=[FileEditor, ""],
    dir=[FileEditor, ""],
    # int=[IntEditor, 0],
    # int2=[Int2Editor, [0,0]],
    # int3=[Int3Editor, [0,0,0]],
    # int8=[IntEditor, 0],
    color=[ColorPicker, [172, 172, 172, 255]],
    # short2=[Float2Editor, [0.0, 0.0]],
    # node=[StringEditor, ""],
    # multi=[StringEditor, ""],
    # doc=[DocumentEditor, ""]
)


def map_widget(attr_type, name, parent=None):
    """
    Map the widget to the attribute type
    :param attr_type:
    :param name:
    :param parent:
    :return:
    """

    widget_type = attr_type.replace(' ', '').lower()
    if widget_type not in WIDGET_MAPPER:
        tp.logger.warning('Invalid Editor Class: "{0}" ("{1}")'.format(attr_type, name))
    else:
        cls = WIDGET_MAPPER.get(widget_type)[0]
        return cls(parent=parent, name=name)
    return
